
#include "../constable.h"
#include "../access_types.h"
#include "fs.h"

/* ---------- fs tree ---------- */

static int node_init( struct tree_s *t )
{ struct fs_s *n= (struct fs_s *)t;
	/* n->inode.dev=0; */
	n->inode.ino=0;
	vs_clear(n->vs);
	if( n->tree.parent->parent==NULL )
		vs_clear(n->recursive_vs);
	else
		vs_set(((struct fs_s *)(n->tree.parent))->recursive_vs,
			n->recursive_vs);
	return(0);
}

static void recursive_add_vs( struct fs_s *f, const vs_t *vs )
{ struct tree_s *t;
	vs_add(vs,f->recursive_vs);
	for(t=f->tree.child;t!=NULL;t=t->next)
		recursive_add_vs((struct fs_s *)t,vs);
	for(t=f->tree.reg;t!=NULL;t=t->next)
		recursive_add_vs((struct fs_s *)t,vs);
}


static int fs_add_vs( struct tree_s *p, int recursive, int which, const vs_t *vs )
{
	if( which != AT_MEMBER )
		return(-1);
	if( recursive==2 )
		vs_add(vs,((struct fs_s *)p)->recursive_vs);
	else if( recursive )
		recursive_add_vs(((struct fs_s *)p),vs);
	else
		vs_add(vs,((struct fs_s *)p)->vs);
	return(0);
}

static tree_type_t tree_type={
			"fs",
			sizeof(struct fs_s),
			NULL,
			node_init,
			&tree_type,
			fs_add_vs,
};

int fs_cinfo;
static int name_offset;
static int name_len;
static int vs_offset;

static int set_handler( struct object_s *o )
{ u_int32_t *o_ci;
  vs_t *vs;
  struct tree_s *t;
  int inh;
	o_ci= (u_int32_t *)(o->data + o->class->cinfo_offset);

	if( (t=(struct tree_s *)(o_ci[fs_cinfo]))==NULL )
		return(-1);
	vs=(vs_t*)(o->data + vs_offset);

	inh=1;
//	for(;t!=NULL;t=t->alt)
//	{	
		vs_add(((struct fs_s *)t)->vs,vs);
		vs_add(((struct fs_s *)t)->recursive_vs,vs);
		object_add_event(o,&(t->event));
		object_add_event(o,&(t->recursive_event));
		if( t->child!=NULL || t->reg!=NULL )
			inh=0;
//	}
	if( !inh )
		object_clr_inherit(o);   /* automaticke dedenie */
	return(0);
}

int iget_handler(struct event_handler_s *h, struct event_context_s *c )
{ u_int32_t *s_ci,*o_ci;
  vs_t *vs;
  	if( c->subject.class!=c->object.class )
		return(-1);
	s_ci= (u_int32_t *)(c->subject.data + c->subject.class->cinfo_offset);
	o_ci= (u_int32_t *)(c->object.data + c->object.class->cinfo_offset);
	if( object_is_same(&(c->subject),&(c->object)) )
	{	/* printf("ZZZ: root\n"); */
		o_ci[fs_cinfo]=s_ci[fs_cinfo]=(u_int32_t)find_one(&global_root,tree_type.name);
	}
	else if( (o_ci[fs_cinfo]=(u_int32_t)find_one((struct tree_s*)(s_ci[fs_cinfo]), c->operation.data + name_offset ))==0 )
		return(-1);

	vs=(vs_t*)(c->object.data + vs_offset);

	if( c->first )
	{	vs_clear(vs);
		object_clear_event(&(c->object));
		object_set_inherit(&(c->object));   /* automaticke dedenie */
	}
	object_do_sethandler(&(c->object));
//printf("ZZZig vs=0x%08x [%04x] (%s) = %s\n",*vs,
//*((u_int16_t*)(c->object.data + c->object.class->event_offset)),
//c->operation.data+name_offset,tree_get_path((struct tree_s*)(o_ci[fs_cinfo])));
	c->result=RESULT_OK;
	return(0);
}

static struct event_handler_s set_handler_d={
			"",
			iget_handler,
			NULL,
};

int fs_init( void )
{ struct medusa_attribute_s *a;
  struct event_type_s *type;
	if( (tree_type.obj_class=get_class_by_name("file"))==NULL )
		return(init_error("fs: Class 'file' don't exist"));
	if( (fs_cinfo=class_alloc_cinfo(tree_type.obj_class))<0 )
		return(init_error("fs: No free cinfo"));
	if( (a=get_attribute(tree_type.obj_class,"vs"))==NULL )
		return(init_error("fs: Class 'file' don't have attribute 'vs'"));
	vs_offset=a->offset;
	if( (type=event_type_findset(tree_type.obj_class))==NULL )
		return(init_error("fs: No set operation!"));
	if( (a=get_attribute(&(type->operation_class),"name"))==NULL )
		return(init_error("fs: set operation don't have attribute 'name'"));
	name_offset=a->offset;
	name_len=a->length;
	if( class_add_sethandler(tree_type.obj_class,set_handler)<0 )
		return(init_error("fs: Can't add sethandler"));
	if( register_event_handler(&set_handler_d,type,ALL_OBJ,ALL_OBJ)<0 )
		return(init_error("fs: Can't register handler"));
	if( register_tree_type(&tree_type)<0 )
		return(init_error("fs: Can't register tree type"));
	return(0);
}

