
#include "../constable.h"
#include "../object.h"
#include "../subject.h"
#include "../space.h"
#include "proc.h"

/* ---------- proc tree ---------- */

static int node_init( struct tree_s *t )
{ struct proc_s *n= (struct proc_s *)t;
  int i;
	for(i=0;i<NR_ACCESS_TYPES;i++)
		vs_clear(n->s.vs[i]);
	if( n->tree.parent->parent==NULL )
		for(i=0;i<NR_ACCESS_TYPES;i++)
			vs_clear(n->s.recursive_vs[i]);
	else
		for(i=0;i<NR_ACCESS_TYPES;i++)
			vs_set(((struct proc_s *)(n->tree.parent))->s.recursive_vs[i], n->s.recursive_vs[i]);
	return(0);
}

static void recursive_add_vs( struct proc_s *f, int which, const vs_t *vs )
{ struct tree_s *t;
	if( which>=0 && which<NR_ACCESS_TYPES )
		vs_add(vs,f->s.recursive_vs[which]);
	for(t=f->tree.child;t!=NULL;t=t->next)
		recursive_add_vs((struct proc_s *)t,which,vs);
	for(t=f->tree.reg;t!=NULL;t=t->next)
		recursive_add_vs((struct proc_s *)t,which,vs);
}

static int proc_add_vs( struct tree_s *p, int recursive, int which, const vs_t *vs )
{
	if( recursive==2 )
	{	if( which>=0 && which<NR_ACCESS_TYPES )
			vs_add(vs,((struct proc_s*)p)->s.recursive_vs[which]);
	}
	else if( recursive )
		recursive_add_vs(((struct proc_s*)p),which, vs);
	else
	{	if( which>=0 && which<NR_ACCESS_TYPES )
			vs_add(vs,((struct proc_s*)p)->s.vs[which]);
	}
	return(0);
}

static tree_type_t tree_type={
			"proc",
			sizeof(struct proc_s),
			NULL,
			node_init,
			&tree_type,
			proc_add_vs,
			};

extern int fs_cinfo;
static int cinfo;
static int vs_offset;
static int vss_offset;
static int vsr_offset;
static int vsw_offset;

static int set_handler_g(struct event_handler_s *h, struct event_context_s *c );

static struct event_handler_s set_handler_d={
			"",
			set_handler_g,
			NULL,
};
static struct event_type_s *e_type=NULL;

static int set_handler( struct object_s *o )
{ u_int32_t *s_ci;
  vs_t *vs,*vss,*vsr,*vsw;
  struct tree_s *t;
  int inh;
	s_ci= (u_int32_t *)(o->data + o->class->cinfo_offset);
	vs=(vs_t*)(o->data + vs_offset);
	vss=(vs_t*)(o->data + vss_offset);
	vsr=(vs_t*)(o->data + vsr_offset);
	vsw=(vs_t*)(o->data + vsw_offset);
//printf("ZZZproc set_handler pred vs=%x vss=%x, vsr=%x, vsw=%x\n",*vs,*vss,*vsr,*vsw);

	if( s_ci[cinfo]==0 )
	{	s_ci[cinfo]=(u_int32_t)find_one(&global_root,tree_type.name);
		vs_clear(vs);		/* !!! ??? ZZZ */
		vs_clear(vss);
		vs_clear(vsr);
		vs_clear(vsw);
		object_clear_event(o);
		object_do_sethandler(o);	/* pre istotu */
		return(0);
	}
	
	t=(struct tree_s *)(s_ci[cinfo]);

	inh=1;
	for(;t!=NULL;t=t->alt)
	{
		vs_add(((struct proc_s *)t)->s.vs[AT_MEMBER],vs);
		vs_add(((struct proc_s *)t)->s.recursive_vs[AT_MEMBER],vs);
		vs_add(((struct proc_s *)t)->s.vs[AT_SEE],vss);
		vs_add(((struct proc_s *)t)->s.recursive_vs[AT_SEE],vss);
		vs_add(((struct proc_s *)t)->s.vs[AT_RECEIVE],vsr);
		vs_add(((struct proc_s *)t)->s.recursive_vs[AT_RECEIVE],vsr);
		vs_add(((struct proc_s *)t)->s.vs[AT_SEND],vsw);
		vs_add(((struct proc_s *)t)->s.recursive_vs[AT_SEND],vsw);
		object_add_event(o,&(t->event));
		object_add_event(o,&(t->recursive_event));
		if( t->child!=NULL || t->reg!=NULL )
			inh=0;
	}
	if( !inh )
		object_add_event(o,&(e_type->mask));
//printf("ZZZproc set_handler vs=%x vss=%x, vsr=%x, vsw=%x\n",*vs,*vss,*vsr,*vsw);
	return(0);
}

static int set_handler_g(struct event_handler_s *h, struct event_context_s *c )
{ u_int32_t *s_ci,*o_ci;
  vs_t *vs,*vss,*vsr,*vsw;
  struct tree_s *t,*n;
  struct space_s *ta;
  vs_t vse[MAX_VS_BITS/32];
  	/* subject - process , object - subor */
	s_ci= (u_int32_t *)(c->subject.data + c->subject.class->cinfo_offset);
	o_ci= (u_int32_t *)(c->object.data + c->object.class->cinfo_offset);

	if( s_ci[cinfo]==0 )
		s_ci[cinfo]=(u_int32_t)find_one(&global_root,tree_type.name);

//printf("ZZZ %s -> ",tree_get_path((struct tree_s*)(s_ci[cinfo])));
	t=(struct tree_s *)(o_ci[fs_cinfo]);
	if( (ta=t->primary_space)==NULL )
		ta=t->recursive_primary_space;
//if( ta!=NULL ) printf("(%s) ",ta->name);
	if( ta!=NULL )
	{	if( (n=find_one((struct tree_s*)(s_ci[cinfo]),ta->name))==NULL )
			return(-1);
	}
	else
	{	if( (n=find_one((struct tree_s*)(s_ci[cinfo])," "))==NULL )
			return(-1);
	}
//printf("%s = ",tree_get_path(n));

	subject_get_vs(vse,AT_ENTER,&(c->subject));
//printf("(%x: %x,%x) ",*vse,((struct proc_s*)n)->s.vs[AT_MEMBER][0],((struct proc_s*)n)->s.recursive_vs[AT_MEMBER][0]);
	if( vs_test(((struct proc_s*)n)->s.vs[AT_MEMBER],vse)
	     || vs_test(((struct proc_s*)n)->s.recursive_vs[AT_MEMBER],vse)
		 )
		s_ci[cinfo]=(u_int32_t)n;
	else
	{	c->result=RESULT_DENY;
		return(0);
	}

//printf("%s\n",tree_get_path((struct tree_s*)(s_ci[cinfo])));

	vs=(vs_t*)(c->subject.data + vs_offset);
	vss=(vs_t*)(c->subject.data + vss_offset);
	vsr=(vs_t*)(c->subject.data + vsr_offset);
	vsw=(vs_t*)(c->subject.data + vsw_offset);

	if( c->first )
	{	vs_clear(vs);
		vs_clear(vss);
		vs_clear(vsr);
		vs_clear(vsw);
		object_clear_event(&(c->subject));
	}
//printf("ZZZproc\n");
	object_do_sethandler(&(c->subject));
	c->result=RESULT_OK;
	return(0);
}

static int proc_get_vs( vs_t *vs, int n, struct object_s *o )
{ u_int32_t *s_ci;
	if( o->class != tree_type.obj_class )
		return(-1);
	s_ci= (u_int32_t *)(o->data + o->class->cinfo_offset);
	if( s_ci[cinfo]==0 )
	{	set_handler(o);
		if( s_ci[cinfo]==0 )
			return(-1);
	}
	vs_add( ((struct proc_s *)(s_ci[cinfo]))->s.vs[n] , vs );
	vs_add( ((struct proc_s *)(s_ci[cinfo]))->s.recursive_vs[n] , vs );
	return(0);
}


int proc_init( void )
{ struct class_s *fs_class;
  struct medusa_attribute_s *a;
	if( (tree_type.obj_class=get_class_by_name("process"))==NULL )
		if( (tree_type.obj_class=get_class_by_name("proc"))==NULL )
			return(init_error("proc: Class 'proc' don't exist"));
	if( (fs_class=get_class_by_name("file"))==NULL )
		return(init_error("proc: Class 'file' don't exist"));

	if( (cinfo=class_alloc_cinfo(tree_type.obj_class))<0 )
		return(init_error("proc: No free cinfo"));
	if( (a=get_attribute(tree_type.obj_class,"vs"))==NULL )
		return(init_error("proc: Class 'proc' don't have attribute 'vs'"));
	vs_offset=a->offset;
	if( (a=get_attribute(tree_type.obj_class,"vss"))==NULL )
		return(init_error("proc: Class 'proc' don't have attribute 'vss'"));
	vss_offset=a->offset;
	if( (a=get_attribute(tree_type.obj_class,"vsr"))==NULL )
		return(init_error("proc: Class 'proc' don't have attribute 'vsr'"));
	vsr_offset=a->offset;
	if( (a=get_attribute(tree_type.obj_class,"vsw"))==NULL )
		return(init_error("proc: Class 'proc' don't have attribute 'vsw'"));
	vsw_offset=a->offset;

	if( (e_type=event_type_find_name("exec",tree_type.obj_class,fs_class,tree_type.obj_class))==NULL )
	   if( (e_type=event_type_find_name("proc_exec",tree_type.obj_class,fs_class,tree_type.obj_class))==NULL )
		return(init_error("proc: No 'proc_exec' operation!"));

	if( class_add_sethandler(tree_type.obj_class,set_handler)<0 )
		return(init_error("proc: Can't add sethandler"));
	if( register_event_handler(&set_handler_d,e_type,ALL_OBJ,ALL_OBJ)<0 )
		return(init_error("proc: Can't register handler"));
	if( subject_register(proc_get_vs)<0 )
		return(init_error("proc: Can't register subject"));
	
	if( register_tree_type(&tree_type)<0 )
		return(init_error("proc: Can't register tree type"));
	return(0);
}

