
#include "../object.h"
#include "../event.h"
#include "comm.h"

#define OFS(x)	( (u_int16_t)(&(((struct m_proc_inf*)0)->x)) )
static struct medusa_attribute_s attr_proc[]={
	{OFS(pid),sizeof(int),MED_TYPE_UNSIGNED|MED_TYPE_PRIMARY_KEY,"pid"},
	{OFS(parent_pid),sizeof(int),MED_TYPE_UNSIGNED|MED_TYPE_READ_ONLY,"ppid"},
	{OFS(child_pid),sizeof(int),MED_TYPE_UNSIGNED|MED_TYPE_READ_ONLY,"child"},
	{OFS(sibling_pid),sizeof(int),MED_TYPE_UNSIGNED|MED_TYPE_READ_ONLY,"sibling"},
	{OFS(pgrp),sizeof(int),MED_TYPE_UNSIGNED,"pgrp"},

	{OFS(uid),sizeof(unsigned short),MED_TYPE_UNSIGNED,"uid"},
	{OFS(euid),sizeof(unsigned short),MED_TYPE_UNSIGNED,"euid"},
	{OFS(suid),sizeof(unsigned short),MED_TYPE_UNSIGNED,"suid"},
	{OFS(fsuid),sizeof(unsigned short),MED_TYPE_UNSIGNED,"fsuid"},
	{OFS(gid),sizeof(unsigned short),MED_TYPE_UNSIGNED,"gid"},
	{OFS(egid),sizeof(unsigned short),MED_TYPE_UNSIGNED,"egid"},
	{OFS(sgid),sizeof(unsigned short),MED_TYPE_UNSIGNED,"sgid"},
	{OFS(fsgid),sizeof(unsigned short),MED_TYPE_UNSIGNED,"fsgid"},
	{OFS(luid),sizeof(unsigned short),MED_TYPE_UNSIGNED|MED_TYPE_READ_ONLY,"luid"},

	{OFS(cap_effective),sizeof(kernel_cap_t),MED_TYPE_BITMAP,"ecap"},
	{OFS(cap_inheritable),sizeof(kernel_cap_t),MED_TYPE_BITMAP,"icap"},
	{OFS(cap_permitted),sizeof(kernel_cap_t),MED_TYPE_BITMAP,"pcap"},

	{OFS(mp.med_vs),4,MED_TYPE_BITMAP,"vs"},
	{OFS(mp.med_vs_s),4,MED_TYPE_BITMAP,"vss"},
	{OFS(mp.med_vs_r),4,MED_TYPE_BITMAP,"vsr"},
	{OFS(mp.med_vs_w),4,MED_TYPE_BITMAP,"vsw"},
	{OFS(mp.med_act),2,MED_TYPE_BITMAP,"event"},
	{OFS(mp.med_user),4,MED_TYPE_BITMAP,"flags"},
	{OFS(mp.cinfo),16,MED_TYPE_UNSIGNED,"cinfo"},
#ifdef CONFIG_MEDUSA_SYSCALL
	{OFS(mp.med_syscall),NR_syscalls/8,MED_TYPE_BITMAP,"syscall"},
#endif
	{0,0,MED_TYPE_END,""},
};
#undef OFS
#define OFS(x)	( (u_int16_t)(&(((struct m_inode_inf*)0)->x)) )
static struct medusa_attribute_s attr_file[]={
	{OFS(dev),sizeof(kdev_t),MED_TYPE_UNSIGNED|MED_TYPE_PRIMARY_KEY|MED_TYPE_READ_ONLY,"dev"},
	{OFS(ino),sizeof(unsigned long),MED_TYPE_UNSIGNED|MED_TYPE_PRIMARY_KEY|MED_TYPE_READ_ONLY,"ino"},
	{OFS(mode),sizeof(__kernel_mode_t),MED_TYPE_BITMAP,"mode"},
	{OFS(uid),sizeof(__kernel_uid_t),MED_TYPE_UNSIGNED,"uid"},
	{OFS(gid),sizeof(__kernel_gid_t),MED_TYPE_UNSIGNED,"gid"},
	{OFS(rdev),sizeof(kdev_t),MED_TYPE_UNSIGNED,"rdev"},
	{OFS(mi.med_vs),4,MED_TYPE_BITMAP,"vs"},
	{OFS(mi.med_act),2,MED_TYPE_BITMAP,"event"},
	{OFS(mi.cinfo),4,MED_TYPE_UNSIGNED,"cinfo"},
#ifdef CONFIG_MEDUSA_LINUX_CAPABILITIES
	{OFS(mi.icap),sizeof(kernel_cap_t),MED_TYPE_BITMAP,"icap"},
	{OFS(mi.pcap),sizeof(kernel_cap_t),MED_TYPE_BITMAP,"pcap"},
	{OFS(mi.ecap),sizeof(kernel_cap_t),MED_TYPE_BITMAP,"ecap"},
#endif
	{0,0,MED_TYPE_END,""},
};
#undef OFS

static struct medusa_class_s class_proc={
	1,sizeof(struct m_proc_inf),"proc"
};

static struct medusa_class_s class_file={
	2,sizeof(struct m_inode_inf),"file"
};

#define OFS(x)	( (u_int16_t)(&(((struct medusa_packet*)0)->u.x)) )
static struct medusa_attribute_s attr_iget[]={
	{sizeof(struct medusa_packet),4096,MED_TYPE_STRING,"name"},
	{0,0,MED_TYPE_END,""},
};
static struct medusa_attribute_s attr_iact[]={
	{OFS(r_iact.act),4,MED_TYPE_UNSIGNED,"action"},
	{OFS(r_iact.info1),4,MED_TYPE_UNSIGNED,"info1"},
	{OFS(r_iact.info2),4,MED_TYPE_UNSIGNED,"info2"},
	{sizeof(struct medusa_packet),4096,MED_TYPE_STRING,"name"},
	{sizeof(struct medusa_packet)+4098,4096,MED_TYPE_STRING,"redirect"},
	{0,0,MED_TYPE_END,""},
};
static struct medusa_attribute_s attr_pact[]={
	{OFS(r_pact.act),4,MED_TYPE_UNSIGNED,"action"},
	{OFS(r_pact.info1),4,MED_TYPE_UNSIGNED,"info1"},
	{OFS(r_pact.info2),4,MED_TYPE_UNSIGNED,"info2"},
	{0,0,MED_TYPE_END,""},
};
static struct medusa_attribute_s attr_trace[]={
	{OFS(r_trace.syscall),2,MED_TYPE_UNSIGNED,"syscall"},
	{OFS(r_trace.syscall),2,MED_TYPE_UNSIGNED,"action"},
	{OFS(r_trace.arg[0]),4,MED_TYPE_UNSIGNED,"arg0"},
	{OFS(r_trace.arg[1]),4,MED_TYPE_UNSIGNED,"arg1"},
	{OFS(r_trace.arg[2]),4,MED_TYPE_UNSIGNED,"arg2"},
	{OFS(r_trace.arg[3]),4,MED_TYPE_UNSIGNED,"arg3"},
	{OFS(r_trace.arg[4]),4,MED_TYPE_UNSIGNED,"arg4"},
	{0,0,MED_TYPE_END,""},
};
static struct medusa_attribute_s attr_getproc[]={
	{0,0,MED_TYPE_END,""},
};
#undef OFS

static struct medusa_optype_s op_iget={
	2,0x01ff,sizeof(struct medusa_packet),"",{2,2},{"parent",""}
};

static struct medusa_optype_s op_i_access={
	2,0,sizeof(struct medusa_packet),"access",{1,2},{"proc","file"}
};
static struct medusa_optype_s op_i_create={
	2,1,sizeof(struct medusa_packet),"create",{1,2},{"proc","file"}
};
static struct medusa_optype_s op_i_link={
	2,2,sizeof(struct medusa_packet),"link",{1,2},{"proc","file"}
};
static struct medusa_optype_s op_i_unlink={
	2,3,sizeof(struct medusa_packet),"unlink",{1,2},{"proc","file"}
};
static struct medusa_optype_s op_i_symlink={
	2,4,sizeof(struct medusa_packet),"symlink",{1,2},{"proc","file"}
};
static struct medusa_optype_s op_i_mkdir={
	2,5,sizeof(struct medusa_packet),"mkdir",{1,2},{"proc","file"}
};
static struct medusa_optype_s op_i_rmdir={
	2,6,sizeof(struct medusa_packet),"rmdir",{1,2},{"proc","file"}
};
static struct medusa_optype_s op_i_mknod={
	2,7,sizeof(struct medusa_packet),"mknod",{1,2},{"proc","file"}
};
static struct medusa_optype_s op_i_rename={
	2,8,sizeof(struct medusa_packet),"rename",{1,2},{"proc","file"}
};
static struct medusa_optype_s op_i_truncate={
	2,11,sizeof(struct medusa_packet),"truncate",{1,2},{"proc","file"}
};
static struct medusa_optype_s op_i_permission={
	2,12,sizeof(struct medusa_packet),"permission",{1,2},{"proc","file"}
};
static struct medusa_optype_s op_i_exec={
	2,13,sizeof(struct medusa_packet),"exec",{1,2},{"proc","file"}
};

//static struct medusa_optype_s op_p_fork={
//	1,0,sizeof(struct medusa_packet),"fork",{1,1},{"parent","child"}
//};
static struct medusa_optype_s op_p_fork={
	1,0,sizeof(struct medusa_packet),"fork",{1,0xffff},{"proc",""}
};
static struct medusa_optype_s op_p_exec1={
	1,1,sizeof(struct medusa_packet),"proc_exec",{1,2},{"proc","file"}
};
static struct medusa_optype_s op_p_exec2={
	1,1,sizeof(struct medusa_packet),"proc_exec",{1,0xffff},{"proc",""}
};
static struct medusa_optype_s op_p_exec3={
	1,0x0301,sizeof(struct medusa_packet),"after_exec",{1,0xffff},{"proc",""}
};
static struct medusa_optype_s op_p_exit={
	1,2,sizeof(struct medusa_packet),"exit",{1,0xffff},{"proc",""}
};
static struct medusa_optype_s op_p_setuid={
	1,3,sizeof(struct medusa_packet),"setuid",{1,0xffff},{"proc",""}
};
static struct medusa_optype_s op_p_kill={
	1,4,sizeof(struct medusa_packet),"kill",{1,1},{"proc","target"}
};
static struct medusa_optype_s op_p_socket={
	1,5,sizeof(struct medusa_packet),"socket",{1,0xffff},{"proc",""}
};
static struct medusa_optype_s op_p_ptrace={
	1,6,sizeof(struct medusa_packet),"socket",{1,1},{"proc","target"}
};
static struct medusa_optype_s op_p_perm={
	1,8,sizeof(struct medusa_packet),"perm",{1,0xffff},{"proc",""}
};
static struct medusa_optype_s op_p_capable={
	1,11,sizeof(struct medusa_packet),"capable",{1,0xffff},{"proc",""}
};
static struct medusa_optype_s op_p_module={
	1,12,sizeof(struct medusa_packet),"module",{1,0xffff},{"proc",""}
};
static struct medusa_optype_s op_p_sexec={
	1,13,sizeof(struct medusa_packet),"sexec",{1,1},{"proc","target"}
};
static struct medusa_optype_s op_p_start={
	1,14,sizeof(struct medusa_packet),"start",{1,0xffff},{"proc","parent"}
};

static struct medusa_optype_s op_trace={
	1,0x04ff,sizeof(struct medusa_packet),"syscall",{1,0xffff},{"proc",""}
};
static struct medusa_optype_s op_getproc={	/* GET_PROC,SET_PROC */
	1,0x05ff,sizeof(struct medusa_packet),"",{1,0xffff},{"proc",""}
};

int comm_vs_size( void )
{
	return(4*8);
}

int comm_init( char *devname, int test )
{
	add_class(&class_proc,attr_proc);
	add_class(&class_file,attr_file);

	event_type_add(&op_iget,attr_iget);

	event_type_add(&op_i_access,attr_iact);
	event_type_add(&op_i_create,attr_iact);
	event_type_add(&op_i_link,attr_iact);
	event_type_add(&op_i_unlink,attr_iact);
	event_type_add(&op_i_symlink,attr_iact);
	event_type_add(&op_i_mkdir,attr_iact);
	event_type_add(&op_i_rmdir,attr_iact);
	event_type_add(&op_i_mknod,attr_iact);
	event_type_add(&op_i_rename,attr_iact);
	event_type_add(&op_i_truncate,attr_iact);
	event_type_add(&op_i_permission,attr_iact);
	event_type_add(&op_i_exec,attr_iact);

	event_type_add(&op_p_fork,attr_pact);
	event_type_add(&op_p_exec1,attr_iact);
	event_type_add(&op_p_exec2,attr_iact);
	event_type_add(&op_p_exec3,attr_iact);
	event_type_add(&op_p_exit,attr_pact);
	event_type_add(&op_p_setuid,attr_pact);
	event_type_add(&op_p_kill,attr_pact);
	event_type_add(&op_p_socket,attr_pact);
	event_type_add(&op_p_ptrace,attr_pact);
	event_type_add(&op_p_perm,attr_pact);
	event_type_add(&op_p_capable,attr_pact);
	event_type_add(&op_p_module,attr_pact);
	event_type_add(&op_p_sexec,attr_pact);
	event_type_add(&op_p_start,attr_pact);

	event_type_add(&op_trace,attr_trace);
	event_type_add(&op_getproc,attr_getproc);

	if( test )
		return(0);
	return( comm_open(devname) );
}

