
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdarg.h>
#include "../constable.h"
#include "../event.h"
#include "comm.h"

static int bit( int mask )
{ int i=0;
	while( (mask>>=1) )
		i++;
	return(i);
}

static void fix_iact( struct med_inode *mi )
{
	if( mi->med_act & 0x8000 )
		mi->med_act|=MIACT_INHERIT;
	else
		mi->med_act&=~MIACT_INHERIT;
	mi->med_act|=MIACT_VALID;
}

static void conv_iget( struct medusa_packet *from, struct medusa_packet *to )
{
	to->u.a_iget.inode = from->u.r_iget.inode;
	fix_iact(&(to->u.a_iget.inode.mi));
//printf("to->u.a_iget.inode.mi.med_act=%x\n",to->u.a_iget.inode.mi.med_act);
}
static void conv_iact( struct medusa_packet *from, struct medusa_packet *to )
{
	to->u.a_iact.proc = from->u.r_iact.proc;
}
static void conv_pact( struct medusa_packet *from, struct medusa_packet *to )
{
	to->u.a_pact.proc = from->u.r_pact.proc;
}
static void conv_trace( struct medusa_packet *from, struct medusa_packet *to )
{
	to->u.a_trace.proc = from->u.r_trace.proc;
	to->u.a_trace.arg[0] = from->u.r_trace.arg[0];
	to->u.a_trace.arg[1] = from->u.r_trace.arg[1];
	to->u.a_trace.arg[2] = from->u.r_trace.arg[2];
	to->u.a_trace.arg[3] = from->u.r_trace.arg[3];
	to->u.a_trace.arg[4] = from->u.r_trace.arg[4];
	to->u.a_trace.retval = -1; /*-EPERM;*/ /* !!! ??? */
}
static void(*conv)(struct medusa_packet *from,struct medusa_packet *to);

static void fix_i_cinfo( struct med_inode *i )
{
//printf("ZZZfix_i_cinfo 1 %04x\n",i->med_act);
	if( !(i->med_act & MIACT_VALID) )
		i->cinfo=0;
//printf("ZZZfix_i_cinfo 2 (%d) (%p)\n",i->cinfo,&(i->cinfo));
}

struct event_type_s *get_event_context( struct event_context_s *c, void *data )
{ struct event_type_s *t;
  struct medusa_packet *mp=data;
  void *s=NULL,*o=NULL;

//if( mp->cmd!=1 )
//printf("ZZZC1 %d\n",mp->cmd);
	if( mp->cmd == MED_IGET )
	{	t=event_type_find(2,0x01ff);
		s=&(mp->u.r_iget.parent);
		fix_i_cinfo(&(mp->u.r_iget.parent.mi));
		fix_i_cinfo(&(mp->u.r_iget.inode.mi));
		o=&(mp->u.r_iget.inode);
		conv=conv_iget;
		if( !(mp->u.r_iget.parent.mi.med_act & MIACT_VALID) &&
			( mp->u.r_iget.parent.dev!=mp->u.r_iget.inode.dev
			|| mp->u.r_iget.parent.ino!=mp->u.r_iget.inode.ino)  )
		{	c->result=RESULT_UNKNOWN;
//printf("IGET: \"%s\"\n",((char*)data)+ ((get_attribute(&(t->operation_class),"name"))->offset) );
//printf("mp->u.r_iget.parent.dev=%d mp->u.r_iget.parent.ino=%d\n",mp->u.r_iget.parent.dev,mp->u.r_iget.parent.ino);
//printf("mp->u.r_iget.inode.dev=%d mp->u.r_iget.inode.ino=%d\n",mp->u.r_iget.inode.dev,mp->u.r_iget.inode.ino);
			return(NULL);
		}
	}
	else if( mp->cmd == MED_IACT && mp->u.r_iact.pact == MPACT_IOP )
	{	c->result=RESULT_UNKNOWN;
//printf("MPACT_IOP???\n");
		return(NULL);
	}
	else if( mp->cmd == MED_IACT && mp->u.r_iact.pact == MPACT_EXEC )
	{	if( mp->u.r_iact.act == 0 )
		{	t=event_type_find(1,1);		/* proc_exec */
//printf("ZZZ: %d proc_exec 0x%04x %s\n",mp->u.r_iact.proc.pid,mp->u.r_iact.proc.mp.med_act,(char*)(&(mp[1])));
			s=&(mp->u.r_iact.proc);
			fix_i_cinfo(&(mp->u.r_iact.inode.mi));
			o=&(mp->u.r_iact.inode);
		}
		else
		{	t=event_type_find(1,0x0301);	/* after_exec */
//printf("ZZZ: %d after_exec\n",mp->u.r_iact.proc.pid);
			s=&(mp->u.r_iact.proc);
			o=NULL;
		}
//printf("ZZZC pid=%d\n",mp->u.r_iact.proc.pid);
		conv=conv_iact;
	}
	else if( mp->cmd == MED_IACT && mp->u.r_iact.pact != MPACT_IOP )
	{	t=event_type_find(2,bit(mp->u.r_iact.act));
		s=&(mp->u.r_iact.proc);
//printf("ZZZC pid=%d\n",mp->u.r_iact.proc.pid);
		fix_i_cinfo(&(mp->u.r_iact.inode.mi));
		o=&(mp->u.r_iact.inode);
		conv=conv_iact;
	}
	else if( mp->cmd == MED_PACT )
	{	t=event_type_find(1,bit(mp->u.r_pact.act));
//printf("ZZZ* %d 0x%04x\n",bit(mp->u.r_pact.act),mp->u.r_pact.act);
		s=&(mp->u.r_pact.proc);
//printf("ZZZC pid=%d\n",mp->u.r_pact.proc.pid);
		o=&(mp->u.r_pact.target);
		if( t->op[1] == NULL )
			o=NULL;
		conv=conv_pact;
	}
	else if( mp->cmd == MED_TRACE )
	{	t=event_type_find(1,0x04ff);
		s=&(mp->u.r_trace.proc);
//printf("ZZZC pid=%d\n",mp->u.r_trace.proc.pid);
		o=NULL;
		conv=conv_trace;
	}
	else if( mp->cmd == MED_GET_PROC )
	{	t=event_type_find(1,0x05ff);
		s=&(mp->u.a_cmd.proc);
		o=NULL;
		conv=NULL;
	}
	else
		t=NULL;
//printf("ZZZC2\n");

 	if( t==NULL )
	{	c->result=RESULT_UNKNOWN;
//printf("ZZZ: Unknown event!!!\n");
		return(NULL);
	}

//if( mp->cmd!=1 )
//printf("ZZZC3\n");
	c->operation.next=&(c->subject);
	c->operation.attr.offset=c->operation.attr.length=0;
	c->operation.attr.type=MED_TYPE_END;
	strncpy(c->operation.attr.name,t->m.name,MIN(MEDUSA_ATTRNAME_MAX,MEDUSA_OPNAME_MAX));
	c->operation.class=&(t->operation_class);
	c->operation.data=(char*)(data);

	c->subject.next=&(c->object);
	c->subject.attr.offset=c->subject.attr.length=0;
	c->subject.attr.type=MED_TYPE_END;
	strncpy(c->subject.attr.name,t->m.op_name[0],MEDUSA_ATTRNAME_MAX);
//printf("ZZZC4\n");
	c->subject.class=t->op[0];
	c->subject.data=s;
	c->object.next=NULL;
	c->object.attr.offset=c->object.attr.length=0;
	c->object.attr.type=MED_TYPE_END;
	strncpy(c->object.attr.name,t->m.op_name[1],MEDUSA_ATTRNAME_MAX);
	c->object.class=t->op[1];
	c->object.data=o;
//printf("ZZZC5\n");
	return(t);
}

static int comm_fd=-1;

#define BUFLEN	4096

char *comm_cmd(struct medusa_packet *msg, char *data)
{
	static char buf[BUFLEN];
	msg->id = 0;
	msg->answer = MED_OK;
	if (data != NULL && msg->data_len <= 0)
		msg->data_len = strlen(data) + 1;
	if (data == NULL)
		msg->data_len = 0;
	if (write(comm_fd, msg, sizeof(struct medusa_packet))
	    != sizeof(struct medusa_packet)) {
		fatal("CMD write error");
		msg->answer = MED_ERR;
		return NULL;
	}
	if (data != NULL) {
		if (write(comm_fd, data, msg->data_len) != msg->data_len) {
			fatal("CMD write data error");
			msg->answer = MED_ERR;
			return NULL;
		}
	}
	if (read(comm_fd, msg, sizeof(struct medusa_packet))
	    != sizeof(struct medusa_packet)) {
		fatal("CMD read error");
		msg->answer = MED_ERR;
		return NULL;
	}
	if (msg->id != 0) {
		fatal("Comunication error!");
		close(comm_fd);
		exit(-1);
	}
	if (msg->data_len > 0) {
		if (msg->data_len >= BUFLEN) {
			fatal("Out of memory!");
			/* FIXME! tell kernel to let it be */
			close(comm_fd);
			exit(-1);
		}
		if (read(comm_fd, buf, msg->data_len) != msg->data_len) {
			fatal("CMD read data error");
			msg->answer = MED_ERR;
			return NULL;
		}
		buf[msg->data_len] = 0;
		return buf;
	}
	return NULL;
}

int medusa_printlog(char *format, ...)
{
	va_list ap;
	int r;
	struct medusa_packet msg;
	char s[4096];

	va_start(ap, format);
#if 0
	if (format[0] == '*')
		format++;
	else
		printf("CONSTABLE LOG: ");
	r = vprintf(format, ap);
#else
	r = vsnprintf(s, 4096, format, ap);
	msg.cmd = MED_PRINTK;
	msg.data_len = 0;
	comm_cmd(&msg, s);
#endif
	va_end(ap);
	return r;
}

int comm_open( char *filename )
{ struct medusa_packet mp;
//printf("ZZZ comm_open return(0);\n"); return(0);
	if( (comm_fd=open(filename,O_RDWR))<0 )
	{	init_error("Can't open %s",filename);
		return(-1);
	}
	mp.id = 0;		/* send supported communcation protocol */
	mp.cmd=MED_VERSION;	/* version number */
	mp.u.r_cmd.arg[0] = 1;	/* only one version is now supported :-) */
	mp.data_len = 0;
	comm_cmd( &mp, NULL );
	if (mp.answer == MED_ERR)
	{	fatal("Unsupported version of communication protocol");
		close(comm_fd);
		return(-1);
	}
	mp.id = 0;
	mp.cmd = MED_RESET;
	mp.data_len = 0;
	comm_cmd( &mp, NULL );
	
	return(0);
}

struct mp_s {
	struct medusa_packet p;
	char data[4098];
	char redirect[4098];
};

static struct mp_s in;

int comm_do( void )
{ int result;
  struct medusa_packet out;

//printf("ZZZ exit()\n"); exit(0);
	if( read(comm_fd,&(in.p),sizeof(in.p)) != sizeof(in.p) )
	{	fatal("Read error!");
		close(comm_fd);
		return(-1);
	}
	/* get the data if available */
	if( in.p.data_len > 0 )
	{	if( in.p.data_len>4096 )
		{	fatal("Out of memory!");
			/* FIXME! tell kernel to let it be */
			close(comm_fd);
			return(-1);
		}
		if( read(comm_fd,in.data,in.p.data_len) != in.p.data_len )
		{	fatal("Data Read error!");
			close(comm_fd);
			return(-1);
		}
		in.data[in.p.data_len]=0;
	}
	else
		in.data[0]=0;
	in.redirect[0]=0;

//printf("ZZZY1\n");
	result=do_event( &in );
//printf("ZZZY2 %d\n",result);

	out.id=in.p.id;
	out.cmd=in.p.cmd;
	out.answer=result;
#ifdef TRANSLATE_RESULT
	switch( result )
	{	case RESULT_ALLOW:
			out.answer=MED_YES;
			break;
		case RESULT_DENY:
			out.answer=MED_NO;
			break;
		case RESULT_SKIP:
			out.answer=MED_SKIP;
			break;
		case RESULT_OK:
			out.answer=MED_OK;
			break;
		default:
			out.answer=MED_ERR;
	}
#endif

	conv(&(in.p),&out);

	if( result>=0 && in.redirect[0]!=0 )
		out.data_len=strlen(in.redirect)+1;
	else
		out.data_len=0;
	if( write(comm_fd,&out,sizeof(out)) != sizeof(out) )
	{	fatal("Write error!");
		close(comm_fd);
		return(-1);
	}
	if( out.data_len!=0 )
	{	if( write(comm_fd,in.redirect,out.data_len) != out.data_len )
		{	fatal("Data Write error!");
			close(comm_fd);
			return(-1);
		}
	}
	return(0);
}

int comm_proc( int pid, struct event_handler_s *handler )
{ struct mp_s in;
  struct medusa_packet out;
  struct event_type_s *t;
  struct event_context_s c;
  void(*conv_a)(struct medusa_packet *from,struct medusa_packet *to);
//printf("comm_proc(%d)\n",pid);
	in.p.cmd = MED_GET_PROC;
	in.p.u.r_cmd.proc.pid = pid;
	comm_cmd(&(in.p), NULL);
	if (in.p.answer != MED_YES || in.p.u.r_cmd.proc.pid != pid) {
		runtime("No such proces %d",pid);
		return(0);
	}
	in.data[0]=0;
	in.redirect[0]=0;
	in.p.cmd=MED_GET_PROC;	/* fix ;-) */
	conv_a=conv;
	if( (t=get_event_context(&c,&in))==NULL )
	{	conv=conv_a;
		return(0);
	};
	conv=conv_a;
	
	c.result=RESULT_UNKNOWN;
	c.first=1;
	c.local_vars=NULL;
	object_do_sethandler(&(c.subject));	/* !!! ??? ZZZ */
	handler->handler(handler,&c);

	out.cmd = MED_SET_PROC;
	out.u.r_cmd.proc = in.p.u.a_cmd.proc;
	comm_cmd(&(out), NULL);
	if (out.answer != MED_YES || out.u.r_cmd.proc.pid != pid) {
		runtime("No such proces %d !",pid);
		return(0);
	}
	return(1);
}


