
#include "../constable.h"
#include "../object.h"
#include "../subject.h"
#include "../language/execute.h"
#include "../language/language.h"
#include "domain.h"

/* ---------- domain tree ---------- */

static int node_init( struct tree_s *t )
{ struct domain_s *n= (struct domain_s *)t;
  int i;
	for(i=0;i<NR_ACCESS_TYPES;i++)
		vs_clear(n->s.vs[i]);
	if( n->tree.parent->parent==NULL )
		for(i=0;i<NR_ACCESS_TYPES;i++)
			vs_clear(n->s.recursive_vs[i]);
	else
		for(i=0;i<NR_ACCESS_TYPES;i++)
			vs_set(((struct domain_s *)(n->tree.parent))->s.recursive_vs[i], n->s.recursive_vs[i]);
	return(0);
}

static void recursive_add_vs( struct domain_s *f, int which, const vs_t *vs )
{ struct tree_s *t;
	if( which>=0 && which<NR_ACCESS_TYPES )
		vs_add(vs,f->s.recursive_vs[which]);
	for(t=f->tree.child;t!=NULL;t=t->next)
		recursive_add_vs((struct domain_s *)t,which,vs);
	for(t=f->tree.reg;t!=NULL;t=t->next)
		recursive_add_vs((struct domain_s *)t,which,vs);
}

static int domain_add_vs( struct tree_s *p, int recursive, int which, const vs_t *vs )
{
	if( recursive==2 )
	{	if( which>=0 && which<NR_ACCESS_TYPES )
			vs_add(vs,((struct domain_s*)p)->s.recursive_vs[which]);
	}
	else if( recursive )
		recursive_add_vs(((struct domain_s*)p),which, vs);
	else
	{	if( which>=0 && which<NR_ACCESS_TYPES )
			vs_add(vs,((struct domain_s*)p)->s.vs[which]);
	}
	return(0);
}

static tree_type_t tree_type={
			"domain",
			sizeof(struct domain_s),
			NULL,
			node_init,
			&tree_type,
			domain_add_vs,
			};

static int cinfo;
static int vs_offset;
static int vss_offset;
static int vsr_offset;
static int vsw_offset;

static int set_handler( struct object_s *o )
{ u_int32_t *s_ci;
  vs_t *vs,*vss,*vsr,*vsw;
  struct tree_s *t;
  int inh;
	s_ci= (u_int32_t *)(o->data + o->class->cinfo_offset);

	if( s_ci[cinfo]==0 )
	{	s_ci[cinfo]=(u_int32_t)find_one(&global_root,tree_type.name);
	}
	
	t=(struct tree_s *)(s_ci[cinfo]);
	vs=(vs_t*)(o->data + vs_offset);
	vss=(vs_t*)(o->data + vss_offset);
	vsr=(vs_t*)(o->data + vsr_offset);
	vsw=(vs_t*)(o->data + vsw_offset);

	inh=1;
	for(;t!=NULL;t=t->alt)
	{
		vs_add(((struct domain_s *)t)->s.vs[AT_MEMBER],vs);
		vs_add(((struct domain_s *)t)->s.recursive_vs[AT_MEMBER],vs);
		vs_add(((struct domain_s *)t)->s.vs[AT_SEE],vss);
		vs_add(((struct domain_s *)t)->s.recursive_vs[AT_SEE],vss);
		vs_add(((struct domain_s *)t)->s.vs[AT_RECEIVE],vsr);
		vs_add(((struct domain_s *)t)->s.recursive_vs[AT_RECEIVE],vsr);
		vs_add(((struct domain_s *)t)->s.vs[AT_SEND],vsw);
		vs_add(((struct domain_s *)t)->s.recursive_vs[AT_SEND],vsw);
		object_add_event(o,&(t->event));
		object_add_event(o,&(t->recursive_event));
		if( t->child!=NULL || t->reg!=NULL )
			inh=0;
	}
/*	if( !inh )
		????????????
*/
	return(0);
}

static int domain_change( struct object_s *o, char *new )
{ u_int32_t *s_ci;
  struct tree_s *t;
  vs_t vs[MAX_VS_BITS/32];
	s_ci= (u_int32_t *)(o->data + o->class->cinfo_offset);
	t=find_one(&global_root,tree_type.name);
	if( (t=find_one(t,new))==NULL )
		return(-1);
	subject_get_vs(vs,AT_ENTER,o);
	if( !vs_test(((struct domain_s*)t)->s.vs[AT_MEMBER],vs)
	     && !vs_test(((struct domain_s*)t)->s.recursive_vs[AT_MEMBER],vs)
		 )
		return(-1);
	s_ci[cinfo]=(u_int32_t)t;

	vs_clear((vs_t*)(o->data + vs_offset));
	vs_clear((vs_t*)(o->data + vss_offset));
	vs_clear((vs_t*)(o->data + vsr_offset));
	vs_clear((vs_t*)(o->data + vsw_offset));
	object_clear_event(o);

	object_do_sethandler(o);
	return(0);
}

static int cmd_domain_change( struct event_context_s *c, int(*getarg)(struct register_s*) )
{ struct register_s r;
	if( getarg(&r)!=1 )
	{	runtime("Change to which domain?");
		return(-1);
	}
	if( (r.attr->type & 0x0f) != MED_TYPE_STRING )
	{	runtime("Change to which domain?");
		return(-1);
	}
	if( domain_change(&(c->subject),r.data)<0 )
	{	runtime("Cannot change domain to %s : Permission denied.",r.data);
		return(-1);
	}
	return(0);
}

static int domain_get_vs( vs_t *vs, int n, struct object_s *o )
{ u_int32_t *s_ci;
	if( o->class != tree_type.obj_class )
		return(-1);
	s_ci= (u_int32_t *)(o->data + o->class->cinfo_offset);
	if( s_ci[cinfo]==0 )
	{	set_handler(o);
		if( s_ci[cinfo]==0 )
			return(-1);
	}
	vs_add( ((struct domain_s *)(s_ci[cinfo]))->s.vs[n] , vs );
	vs_add( ((struct domain_s *)(s_ci[cinfo]))->s.recursive_vs[n] , vs );
	return(0);
}

int domain_init( void )
{ struct medusa_attribute_s *a;
	if( (tree_type.obj_class=get_class_by_name("process"))==NULL )
		if( (tree_type.obj_class=get_class_by_name("proc"))==NULL )
			return(init_error("domain: Class 'proc' don't exist"));

	if( (cinfo=class_alloc_cinfo(tree_type.obj_class))<0 )
		return(init_error("domain: No free cinfo"));
	if( (a=get_attribute(tree_type.obj_class,"vs"))==NULL )
		return(init_error("domain: Class 'proc' don't have attribute 'vs'"));
	vs_offset=a->offset;
	if( (a=get_attribute(tree_type.obj_class,"vss"))==NULL )
		return(init_error("domain: Class 'proc' don't have attribute 'vss'"));
	vss_offset=a->offset;
	if( (a=get_attribute(tree_type.obj_class,"vsr"))==NULL )
		return(init_error("domain: Class 'proc' don't have attribute 'vsr'"));
	vsr_offset=a->offset;
	if( (a=get_attribute(tree_type.obj_class,"vsw"))==NULL )
		return(init_error("domain: Class 'proc' don't have attribute 'vsw'"));
	vsw_offset=a->offset;

	if( class_add_sethandler(tree_type.obj_class,set_handler)<0 )
		return(init_error("domain: Can't add sethandler"));
	if( subject_register(domain_get_vs)<0 )
		return(init_error("domain: Can't register subject"));

	if( register_tree_type(&tree_type)<0 )
		return(init_error("domain: Can't register tree type"));

	lex_addkeyword("enter_domain",Tbuildin,(unsigned long)cmd_domain_change);
	return(0);
}

