
#include "constable.h"
#include "event.h"
#include "tree.h"

struct event_type_s *get_event_context( struct event_context_s *c, void *data );

int event_mask_clear( event_mask_t *e )
{
	*e=0;
	return(0);
}

int event_mask_setbit( event_mask_t *e, int b )
{
	if( b<0 || b>=sizeof(event_mask_t)*8 )
		return(-1);
	*e|= 1<<b ;
	return(0);
}

int event_mask_clrbit( event_mask_t *e, int b )
{
	if( b<0 || b>=sizeof(event_mask_t)*8 )
		return(-1);
	*e&= ~(1<<b) ;
	return(0);
}

int event_mask_and( event_mask_t *e, event_mask_t *f )
{
	*e &= *f;
	return(0);
}

int event_mask_or( event_mask_t *e, event_mask_t *f )
{
	*e |= *f;
	return(0);
}

int event_mask_sub( event_mask_t *e, event_mask_t *f )
{
	*e &= ~(*f);
	return(0);
}

struct event_type_s *global_event_types=NULL;
static struct event_type_s *last_event_type=NULL;

static struct event_type_s **event_type_hash=NULL;
static int event_type_hash_maxclass,event_type_hash_maxop,event_type_hash_maxv;

static struct event_type_s *event_type_slow_find( u_int16_t classid, u_int16_t opid )
{ struct event_type_s *p;
	for(p=global_event_types;p!=NULL;p=p->next)
		if( p->m.opid == opid && p->m.classid == classid )
			return(p);
	return(NULL);
}

int event_type_add( struct medusa_optype_s *m, struct medusa_attribute_s *a )
{ struct event_type_s *p;
  int l;
	for(l=0;a[l].type!=MED_TYPE_END;l++);
	l++;
	l*=sizeof(struct medusa_attribute_s);
	if( (p=malloc(sizeof(struct event_type_s)+l))==NULL )
		return(-1);
	memcpy(&(p->m),m,sizeof(struct medusa_optype_s));
	p->operation_class.next=NULL;
	p->operation_class.cinfo_offset=p->operation_class.cinfo_size=0;
	p->operation_class.cinfo_mask=0;
	p->operation_class.set=NULL;
	p->operation_class.event_offset=p->operation_class.event_size=0;
	p->operation_class.m.classid=0;
	p->operation_class.m.size=p->m.size;
	p->operation_class.m.name[0]=0;
	memcpy(p->operation_class.attr,a,l);
	p->op[0]=get_class_by_id(p->m.op_class[0]);
	p->op[1]=get_class_by_id(p->m.op_class[1]);
	p->object=get_class_by_id(p->m.classid);
	event_mask_clear(&(p->mask));
	event_mask_setbit(&(p->mask),(p->m.opid)&0x00ff);
	p->handlers_hash=NULL;
	p->next=NULL;
	p->alt=event_type_slow_find(p->m.classid,p->m.opid);
	if( last_event_type==NULL )
		global_event_types=p;
	else	last_event_type->next=p;
	last_event_type=p;
	if( event_type_hash!=NULL )
		free(event_type_hash);
	event_type_hash=NULL;
	return(0);
}

#define	EVTH(c,o)	event_type_hash[(c)*event_type_hash_maxop*event_type_hash_maxv+((o)>>8)*event_type_hash_maxop+((o)&0x00ff)]

int event_type_hash_create( void )
{ struct event_type_s *p;
	if( event_type_hash!=NULL )
		free(event_type_hash);
	event_type_hash_maxclass=event_type_hash_maxop=event_type_hash_maxv=0;
	for(p=global_event_types;p!=NULL;p=p->next)
	{	if( p->m.classid > event_type_hash_maxclass )
			event_type_hash_maxclass=p->m.classid;
		if( ((p->m.opid)&0x00ff) > event_type_hash_maxop )
			event_type_hash_maxop=((p->m.opid)&0x00ff);
		if( ((p->m.opid)>>8) > event_type_hash_maxv )
			event_type_hash_maxv=((p->m.opid)>>8);
	}
	event_type_hash_maxclass++;
	event_type_hash_maxop++;
	event_type_hash_maxv++;
	if( (event_type_hash=malloc(event_type_hash_maxclass*event_type_hash_maxop*event_type_hash_maxv*sizeof(struct event_type_s *)))==NULL )
		return(-1);
	memset(event_type_hash,0,event_type_hash_maxclass*event_type_hash_maxop*event_type_hash_maxv*sizeof(struct event_type_s *));
	for(p=global_event_types;p!=NULL;p=p->next)
		if( p->alt==NULL && EVTH(p->m.classid,p->m.opid)==NULL )
			EVTH(p->m.classid,p->m.opid)=p;
	return(0);
}

struct event_type_s *event_type_find( u_int16_t classid, u_int16_t opid )
{
	if( event_type_hash==NULL )
		event_type_hash_create();
	if( classid>=event_type_hash_maxclass || (opid&0x00ff)>=event_type_hash_maxop || (opid>>8)>=event_type_hash_maxv )
//{ printf("ZZZ: event_type_find out %d>=%d %d>=%d\n",classid,event_type_hash_maxclass,opid,event_type_hash_maxop);
		return(NULL);
//}
//printf("ZZZ: event_type_find %p\n",event_type_hash[classid*event_type_hash_maxop+opid]);
	return(EVTH(classid,opid));
}

struct event_type_s *event_type_find_name( char *name, struct class_s *s, struct class_s *o, struct class_s *ob )
{ struct event_type_s *t;
	for(t=global_event_types;t!=NULL;t=t->next)
		if( !strncmp(t->m.name,name,MEDUSA_OPNAME_MAX)
		    && t->op[0]==s && t->op[1]==o && (ob==NULL || t->object==ob) )
			break;
	return(t);
}

struct event_type_s *event_type_findset( struct class_s *c )
{ struct event_type_s *t;
	for(t=global_event_types;t!=NULL;t=t->next)
		if( t->m.name[0]==0 && t->op[0]==c && t->op[1]==c )
			break;
	return(t);
}

struct event_hadler_hash_s *evhash_add( struct event_hadler_hash_s **hash, struct event_handler_s *handler )
{ struct event_hadler_hash_s *l;
	if( (l=malloc(sizeof(struct event_hadler_hash_s)))==NULL )
	{	errstr=Out_of_memory;
		return(NULL);
	}
	for(;*hash!=NULL;hash=&((*hash)->next))
	{	if( (*hash)->handler==handler )
		{	free(l);
			return(*hash);
		}
	}
	l->next=NULL;
	l->handler=handler;
	*hash=l;
	return(l);
}


int register_event_handler( struct event_handler_s *h, struct event_type_s *type, vs_t *subject_vs, vs_t *object_vs )
{ struct event_hadler_hash_s *l;
  struct event_type_s *ftype=type;
  	while( ftype->alt )	
		ftype=ftype->alt;
	if( (l=evhash_add(&(ftype->handlers_hash),h))==NULL )
		return(-1);
	l->type=type;
	if( subject_vs==ALL_OBJ || (subject_vs==NULL && ftype->op[0]!=NULL) )
		vs_fill(l->subject_vs);
	else if( subject_vs )
		vs_set(subject_vs,l->subject_vs);
	else	vs_clear(l->subject_vs);
	if( object_vs==ALL_OBJ || (object_vs==NULL && ftype->op[1]!=NULL) )
		vs_fill(l->object_vs);
	else if( object_vs )
		vs_set(object_vs,l->object_vs);
	else	vs_clear(l->object_vs);
//printf("Zaregistrovane %p [%s]\n",l,l->type->m.name);
	return(0);
}


int evaluate_result( int old, int new )
{
/*
	if( new==RESULT_ALLOW && (old==RESULT_UNKNOWN || old==RESULT_OK) )
		return(new);
	else if( new==RESULT_DENY )
		return(new);
	else if( new==RESULT_SKIP && old!=RESULT_DENY )
		return(new);
	else if( new==RESULT_OK && old==RESULT_UNKNOWN )
		return(new);
*/
	if( old==RESULT_DENY || new==RESULT_DENY )
		return(RESULT_DENY);
	if( old==RESULT_UNKNOWN || new==RESULT_SKIP )
		return(new);
	if( old==RESULT_OK && new==RESULT_ALLOW )
		return(new);
	return(old);
}

int do_event( void *data )
{ struct event_handler_s *h;
  struct event_hadler_hash_s *hh;
  struct event_context_s c,ct;
  struct event_type_s *t;
  vs_t *v_s,*v_o;
  
//printf("ZZZZ1\n");
	if( (t=get_event_context(&c,data))==NULL )
		return(c.result);
  	while( t->alt )
		t=t->alt;
//printf("ZZZZ2\n");

//if( t->m.name[0]!=0 /* nie iget na subory */ )
//{
//printf("ZZZ: event: op=%s ",t->m.name);
//if( c.subject.class!=NULL )
//  printf("subject=%s ",c.subject.class->m.name);
//else
//  printf("subject=NONE ");
//if( c.object.class!=NULL )
//  printf("object=%s ",c.object.class->m.name);
//else
//  printf("object=NONE ");
//printf("\n");
//}
 
	c.result=RESULT_UNKNOWN;
	c.first=1;
	c.local_vars=NULL;
	object_do_sethandler(&(c.subject));	/* !!! ??? ZZZ */
	if( c.subject.class!=NULL && c.subject.class->vs_size>0 )
		v_s=(vs_t*)(c.subject.data+c.subject.class->vs_offset);
	else
		v_s=NULL;
	if( c.object.class!=NULL && c.object.class->vs_size>0 )
		v_o=(vs_t*)(c.object.data+c.object.class->vs_offset);
	else
		v_o=NULL;

	evhash_foreach(hh,t->handlers_hash)
	{	h=hh->handler;
//printf("testujem handler %p v_s=%p v_o=%p ic_s=%d ic_o=%d\n",hh,v_s,v_o,
//vs_isclear(hh->subject_vs),vs_isclear(hh->object_vs));
		if( (v_s==NULL) != (vs_isclear(hh->subject_vs)!=0) )
			continue;
		if( (v_o==NULL) != (vs_isclear(hh->object_vs)!=0) )
			continue;
//printf("hh->subject_vs=0x%08x ? v_s=0x%08x\n",hh->subject_vs[0],v_s[0]);
		if( v_s!=NULL && !vs_isfull(hh->subject_vs) )
		{	if( !vs_test(hh->subject_vs,v_s) )
				continue;
		}
//printf("hh->object_vs=0x%08x ? v_o=0x%08x\n",hh->object_vs[0],v_o[0]);
		if( v_o!=NULL && !vs_isfull(hh->object_vs) )
		{	if( !vs_test(hh->object_vs,v_o) )
				continue;
		}
//printf("handler\n");
		ct=c;
		ct.result=RESULT_UNKNOWN;
		ct.local_vars=h->local_vars;
		h->handler(h,&ct);
		c.result=evaluate_result(c.result,ct.result);
		c.first=0;
	}
//if( c.first ) printf("do_event: No event executed! [%s]\n",t->m.name);
	return(c.result);
}

