
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sched.h>
#include <signal.h>
#include "event.h"

#ifndef MEDUSA_INITNAME
#define MEDUSA_INITNAME	"/sbin/init"
#endif

int comm_init( char *devname, int test );
int procmem_init( void );
int vs_init( int max_vs );
int fs_init( void );
int proc_init( void );
int domain_init( void );
int language_init( char *filename );
int execute_init( void );
int cmds_init( void );

int comm_vs_size( void );
int comm_do( void );

int language_do( void );

extern struct event_handler_s *function_init;
int comm_proc( int pid, struct event_handler_s *handler );


#ifdef RBAC
int rbac_init( void );
int rbac_load( char *userfile, char *conffile );
#endif


int init( char *devname, char *filename, int test )
{
	if( comm_init(devname,test)<0 )
		return(-1);
	if( vs_init(comm_vs_size())<0 )
		return(-1);
	if( cmds_init()<0 )
		return(-1);
	if( procmem_init()<0 )
		return(-1);
	if( fs_init()<0 )
		return(-1);
	if( proc_init()<0 )
		return(-1);
	if( domain_init()<0 )
//		return(-1);
	;
#ifdef RBAC
	if( rbac_init()<0 )
//		return(-1);
	;
#endif
	if( execute_init()<0 )
		return(-1);
	if( language_init(filename)<0 )
		return(-1);
	if( language_do()<0 )
		return(-1);
//printf("ZZZ: All initialized\n");
	return(0);
}

int usage( char *me )
{
	fprintf(stderr,"usage: %s [-t] [<config. file>] [-d <device>]"
#ifdef RBAC
		" [-r <rbac config. file>]"
#endif
		"\n",me);
	return(0);
}

static int run_init( int argc, char *argv[] )
{ int i;
//  int status;

	switch ((i = fork()))
	{
		case -1:
			return(0);
		case 0:
			return(1);
		default:
			pause();
//			waitpid(i, &status, 0);
//			if (!WIFEXITED(status))
//				exit(-1);
			argv[0] = MEDUSA_INITNAME;
			execvp(argv[0], argv);
			exit(-1);
	}
	return(0);
}

int main( int argc, char *argv[] )
{ char *conf_name="/etc/medusa.conf";
  char *devname="/dev/medusa";
#ifdef RBAC
  char *rbac_conf="/etc/rbac.conf";
#endif
  struct sched_param schedpar;
  int a;
  int test=0;
  int kill_init=0;

	if( getpid() <= 1 )
		kill_init=run_init(argc,argv);

	for(a=1;a<argc;a++)
	{
		if( argv[a][0]=='-' )
		{	if( argv[a][1]=='t' )
				test=1;
			else if( argv[a][1]=='d' && a+1<argc )
			{	a++;
				devname=argv[a];
			}
#ifdef RBAC
			else if( argv[a][1]=='r' && a+1<argc )
			{	a++;
				rbac_conf=argv[a];
			}
#endif
			else
				return(usage(argv[0]));
		}
		else
		{	conf_name=argv[a];
		}
	}

	if( init(devname,conf_name,test)<0 )
		return(-1);
#ifdef RBAC
	if( rbac_load("/etc/passwd",rbac_conf)<0 )
		return(-1);
#endif
	if( test )
		return(0);
	setsid();
	schedpar.sched_priority = sched_get_priority_max(SCHED_FIFO);
	sched_setscheduler(0, SCHED_FIFO, &schedpar);
	if( function_init!=NULL )
		comm_proc(getpid(),function_init);
	if( kill_init )
		kill(1,SIGHUP);
	while( comm_do()>=0 );
	return(-1);
}

