
#include "execute.h"
#include "../constable.h"

int r_int( struct register_s *r )
{ u_int32_t *p;
	if( r->data==r->buf )
		p=((u_int32_t*)(r->data));
	else	p=((u_int32_t*)(r->data+r->attr->offset));
	if( (r->attr->type & 0x0f)==MED_TYPE_SIGNED )
		return((int)(*p));
	if( (r->attr->type & 0x0f)==MED_TYPE_UNSIGNED )
		return((int)(*p));
	runtime("r_int from non integer");
	return(0);
}

void r_imm( struct register_s *r )
{ int n,m;
  u_int32_t x;
	if( r->data==r->buf )
		return;
	n=r->attr->length;
	m= (n+3) >> 2;
	if( n>MAX_REG_SIZE )
		fatal("Variable too long");
	if( m<<2 != n )
		((u_int32_t*)(r->buf))[m-1]=0;
	memcpy(r->buf,r->data + r->attr->offset,n);
	if( m<<2 != n && (r->attr->type & 0x0f)==MED_TYPE_SIGNED )
	{	x=((u_int32_t*)(r->buf))[m-1];
		switch( n % 4 )
		{	case 1:	if( x & 0x00000080 )
					x|=0xffffff00;
				break;
			case 2:	if( x & 0x00008000 )
					x|=0xffff0000;
				break;
			case 3:	if( x & 0x00800000 )
					x|=0xff000000;
				break;
		}
		((u_int32_t*)(r->buf))[m-1]=x;
	}
	r->data=r->buf;
}

int strnlen( char *s, int n )
{ int i=0;
	while( i<n && s[i]!=0 )
		i++;
	return(i);
}

void r_sto( struct register_s *v, struct register_s *d )
{ int nv,nd;
  u_int8_t type;
	if( v->data==v->buf )
	{	runtime("Invalid lvalue");
if( (d->attr->type & 0x0f)==MED_TYPE_STRING )
//printf("ZZZ: r_sto: str=\"%s\"\n",d->data);
		return;
	}
	if( v->attr->type & MED_TYPE_READ_ONLY )
	{	runtime("Changing read-only varible");
		return;
	}
	type=v->attr->type&0x0f;
	if( type==MED_TYPE_UNSIGNED && (d->attr->type & 0x0f)==MED_TYPE_SIGNED )
		type=MED_TYPE_UNSIGNED;
	else if( type==MED_TYPE_SIGNED && (d->attr->type & 0x0f)==MED_TYPE_UNSIGNED )
		type=MED_TYPE_UNSIGNED;
	else if( type==MED_TYPE_BITMAP && (d->attr->type & 0x0f)==MED_TYPE_SIGNED )
		type=MED_TYPE_SIGNED;
	else if( type==MED_TYPE_BITMAP && (d->attr->type & 0x0f)==MED_TYPE_UNSIGNED )
		type=MED_TYPE_UNSIGNED;
	else if( (v->attr->type & 0x0f) != (d->attr->type & 0x0f) )
	{	runtime("Illegal type conversion [%1x->%1x]",(d->attr->type & 0x0f),(v->attr->type & 0x0f));
		return;
	}
	if( d->data!=d->buf )
		r_imm(d);
	
	nv=v->attr->length;
	nd=d->attr->length;
	if( nv!=nd )
	{	switch( type )
		{	
			case MED_TYPE_STRING:
				if( nv>nd )
				{	memcpy( v->data + v->attr->offset, d->data , nd );
					(v->data + v->attr->offset)[nd]=0;
				}
				else
				{	nd=strnlen(d->data,nv);
					memcpy( v->data + v->attr->offset, d->data , nd );
				}
				break;
			case MED_TYPE_UNSIGNED:
			case MED_TYPE_BITMAP:
				if( nv > nd )
					memset( v->data + v->attr->offset, 0 , nv );
				memcpy( v->data + v->attr->offset, d->data , MIN(nd,nv) );
				break;
			case MED_TYPE_SIGNED:
				if( nv > nd )
				{	if( ((u_int32_t*)(d->data))[((nd+3)>>2)-1] & 0x80000000 )
						memset( v->data + v->attr->offset, 0xff , nv );
					else
						memset( v->data + v->attr->offset, 0 , nv );
				}
				memcpy( v->data + v->attr->offset, d->data , MIN(((nd+3)>>2)<<2,nv) );
				break;
			case MED_TYPE_END:
				runtime("Class of object don't match");
				break;
		}
	}
	else
		memcpy( v->data + v->attr->offset, d->data , nv );
}

void r_resize( struct register_s *v, int size )
{ int nv;
  u_int8_t type;
	if( v->data!=v->buf )
		r_imm(v);
	type=v->attr->type&0x0f;
	
	nv=v->attr->length;
	if( nv >= size )
		return;
	switch( type )
	{	
		case MED_TYPE_STRING:
			break;
		case MED_TYPE_UNSIGNED:
			nv= (nv+3) & ~0x03;
			if( nv>= size )
				break;
		case MED_TYPE_BITMAP:
			memset( v->data + nv, 0 , size-nv );
			break;
		case MED_TYPE_SIGNED:
			nv= (nv+3) & ~0x03;
			if( nv>= size )
				break;
			if( ((u_int32_t*)(v->data))[(nv>>2)-1] & 0x80000000 )
				memset( v->data + nv, 0xff , size-nv );
			else
				memset( v->data + nv, 0 , size-nv );
			break;
		case MED_TYPE_END:
			runtime("Object can't be resized");
			break;
	}
}

