
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include "../constable.h"
#include "error.h"
#include "language.h"

extern compiler_class_t *global_compiler;

char *errstr=NULL;
char *Out_of_memory="Out of memory";

int error( const char *fmt, ... )
{ va_list ap;
  char buf[4096];

	sprintf(buf,"%.40s [%d,%d]: Error ",
		global_compiler->lex->filename,
		global_compiler->lex->row,
		global_compiler->lex->col);

	va_start(ap,fmt);
	vsnprintf(buf+strlen(buf), 4000, fmt, ap);
	va_end(ap);
	sprintf(buf+strlen(buf),"\n");
	write(1,buf,strlen(buf));
	global_compiler->err->errors++;
	return(0);
}

int warning( const char *fmt, ... )
{ va_list ap;
  char buf[4096];

	sprintf(buf,"%.40s [%d,%d]: Warning ",
		global_compiler->lex->filename,
		global_compiler->lex->row,
		global_compiler->lex->col);

	va_start(ap,fmt);
	vsnprintf(buf+strlen(buf), 4000, fmt, ap);
	va_end(ap);
	sprintf(buf+strlen(buf),"\n");
	write(1,buf,strlen(buf));
	global_compiler->err->warnings++;
	return(0);
}



static char *sym2str( sym_t sym )
{ lextab_t *l=keywords;
  static char buf[16];
	if( sym==TEND )	return("End of file");
	if( sym==T_id )	return("identifier");
	if( sym==T_str )	return("string");
	if( sym==T_num )	return("constant");
	if( sym==Taccess )	return("access type");
	while(l->keyword!=NULL)
	{	if( l->sym==sym )
			return(l->keyword);
		l++;
	}
	l=keywords2;
	while(l->keyword!=NULL)
	{	if( l->sym==sym )
			return(l->keyword);
		l++;
	}
	l=operators;
	while(l->keyword!=NULL)
	{	if( l->sym==sym )
			return(l->keyword);
		l++;
	}
	sprintf(buf,"?%04x?",sym);
	return(buf);
}
static sym_t err_warning( struct compiler_err_class *this, sym_t errsym, sym_t info )
{
	if( errsym==TEND )	return(0);
	if( errsym==END && (info&TYP)==E )
		warning("Lexical warning %04x",info&~TYP);
	else if( (errsym&TYP)==W && info==END )
		warning("%04x",errsym&~TYP);
	else	warning("%04x %04x",errsym,info);
	return(0);
}
static sym_t err_error( struct compiler_err_class *this, sym_t errsym, sym_t info )
{
	if( errsym==TEND )	return(0);
	if( errsym==END && info==eLEXERR )
		error("Lexical error");
	else if( errsym==END && (info&TYP)==E )
		error("Lexical error %04x",info&~TYP);
	else if( errsym==END && ((info&TYP)==T || info==TEND) )
		error("Unexpected %s",sym2str(info));
	else if( (errsym&TYP)==T && ((info&TYP)==T || info==TEND) )
		error("Missing %s",sym2str(errsym));
	else if( errsym==(E|1) )
		error("%s",errstr);
	else if( (errsym&TYP)==E && info==END )
		error("%04x",errsym&~TYP);
	else	error("%04x %04x",errsym,info);
	return(0);
}

static void err_destroy( struct compiler_err_class *this )
{	return;
}

struct compiler_err_class s_error={
		err_destroy,
		0,
		err_warning,err_error,0,0
};

