
#include "execute.h"
#include "language.h"
#include "../constable.h"

int comm_proc( int pid, struct event_handler_s *handler );
int force_do( u_int32_t fc, struct event_context_s *c, int(*getarg)(struct register_s*) );

static u_int32_t *stack=NULL;
static int stack_size;
static int stack_pos;
static int base_stack_pos=0;

#define	LTI	0
#define	LTS	1

static void push( u_int32_t data )
{
	if( stack_pos>=stack_size )
	{	stack_size= stack_pos + 256;
		if( (stack=realloc(stack,stack_size*sizeof(u_int32_t)))==NULL )
		{	fatal("Stack overrun");
		}
	}
	stack[stack_pos++]=data;
}

static u_int32_t pop( void )
{
	if( stack_pos <= 0 )
	{	fatal("Stack underrun");
	}
	return( stack[--stack_pos] );
}

static struct medusa_attribute_s attr_int={ 0,4,MED_TYPE_SIGNED|MED_TYPE_READ_ONLY,"" };
static struct medusa_attribute_s attr_str={ 0,MAX_REG_SIZE,MED_TYPE_STRING|MED_TYPE_READ_ONLY,"" };

void r_pop( struct register_s *r )
{ u_int32_t x;
  int i,n;
	x=pop();
	switch( x )
	{	case LTI:
			r->attr=&attr_int;
			r->data=r->buf;
			*((u_int32_t*)(r->data))=pop();
			return;
		case LTS:
//			r->attr=&attr_str;
			r->tmp_attr=attr_str;
			r->data=(char *)pop();
			r->tmp_attr.length=strlen( (char *)(r->data) )+1;
			r->attr=&(r->tmp_attr);
			return;
		default:
			r->attr=(struct medusa_attribute_s *)x;
			x=pop();
			if( x!=0 )
			{	r->data=(char*)x;
				return;
			}
			r->data=r->buf;
			n= (r->attr->length + 3) >> 2;
			for(i=0;i<n;i++)
				((u_int32_t*)(r->data))[i]=pop();
			return;
	}
}

void r_push( struct register_s *r )
{ int n;
	if( r->attr==&attr_int )
	{	push( *((u_int32_t*)(r->data)) );
		push( LTI );
	}
	else if( r->attr->type==attr_str.type && r->data!=r->buf )
	{	push( (u_int32_t)(r->data) );
		push( LTS );
	}
	else if( r->data!=r->buf )
	{	push( (u_int32_t)(r->data) );
		push( (u_int32_t)(r->attr) );
	}
	else
	{	n= (r->attr->length + 3) >> 2;
		for(n--;n>=0;n--)
			push( ((u_int32_t*)(r->data))[n] );
		push( 0 );
		push( (u_int32_t)(r->attr) );
	}
}

static int fn_getsval_n=0;
static int fn_getsval( struct register_s *r )
{ int z_stack_pos,i;
	if( r==NULL )
		return(fn_getsval_n);
	if( fn_getsval_n<=0 )
		return(0);
	z_stack_pos=stack_pos;
	for(i=fn_getsval_n;i>0;i--)
	{	r_pop(r);
	}
	stack_pos=z_stack_pos;
	fn_getsval_n--;
	r_imm(r);
	return(1);
}

static int fn_getargsval( struct register_s *r, int n )
{ int z_stack_pos;
	z_stack_pos=stack_pos;
	stack_pos=base_stack_pos;
	pop();
	pop();
	for(;n>0;n--)
	{	r_pop(r);
	}
	stack_pos=z_stack_pos;
	r_imm(r);
	return(1);
}

static struct register_s r0,r1;

int execute_handler( struct event_handler_s *h, struct event_context_s *c )
{ u_int32_t *p,cmd=oNOP,x;
  struct medusa_attribute_s *a;

	c->result=RESULT_OK;

	p=(u_int32_t *)(h->data);
for(;;)
{	cmd=*p++;
	switch( cmd )
	{
	case oNOP:	break;
	case oLTI:	push(LTI);
			break;
	case oLTS:	push(LTS);
			break;
	case oLDI:	push(*p++);
			break;
	case oLD0:	push(0);
			push(LTI);
			break;
	case oLD1:	push(1);
			push(LTI);
			break;
	case oDUP:	r_pop(&r0);
			r_push(&r0);
			r_push(&r0);
			break;
	case oSWP:	r_pop(&r0);
			r_pop(&r1);
			r_push(&r0);
			r_push(&r1);
			break;
	case oDEL:	r_pop(&r0);
			break;
	case oDEn:	x=*p++;
			r_pop(&r1);
			while( x )
			{	r_pop(&r0);
				x--;
			}
			r_push(&r1);
			break;
	case oIMM:	r_pop(&r0);
			r_imm(&r0);
			r_push(&r0);
			break;
	case oSTO:	r_pop(&r0);
			r_pop(&r1);
			r_sto(&r1,&r0);
			r_push(&r1);
			break;
	case oLDC:	x=*p++;
			load_constant(&r0,x,(char*)(*p++));
			r_push(&r0);
			break;
	case oLDV:	if( (x=data_load_var(c,(char*)(*p++),NULL,&a))==0 )
			{	push(x);
				push(LTI);
			}
			else
			{	push(x);
				push((u_int32_t)a);
			}
			break;
	case oLDS:	x=*p++;
			if( (x=data_load_var(c,(char*)x,(char*)(*p++),&a))==0 )
			{	push(x);
				push(LTI);
			}
			else
			{	push(x);
				push((u_int32_t)a);
			}
			break;
	case oADD:
	case oSUB:
	case oMUL:
	case oDIV:
	case oMOD:
	case oSHL:
	case oSHR:
	case oLT:
	case oGT:
	case oLE:
	case oGE:
	case oEQ:
	case oNE:
	case oOR:
	case oAND:
	case oXOR:
			r_pop(&r0);
			r_pop(&r1);
			do_bin_op(cmd,&r1,&r0);
			r_push(&r1);
			break;
	case oNOT:	r_pop(&r0);
			r_not(&r0);
			r_push(&r0);
			break;
	case oNEG:	r_pop(&r0);
			r_imm(&r0);
			r_neg(&r0);
			r_push(&r0);
			break;

	case oJR:	p+= *p++;
			break;
	case oJIZ:	r_pop(&r0);
			if( ! r_nz(&r0) )
				p+= *p++;
			else	p++;
			break;
	case oJNZ:	r_pop(&r0);
			if( r_nz(&r0) )
				p+= *p++;
			else	p++;
			break;
	case oJSR:	x=*p++;
//printf("%p JSR %x\n",p-2,x);
			push((u_int32_t)p);
				push(base_stack_pos);
				base_stack_pos=stack_pos;
			p=(u_int32_t*)x;
			break;
	case oRET:	
//printf("%p RET\n",p-1);
			r_pop(&r0);
			if( stack_pos <= 0 )
			{	r_imm(&r0);
				c->result=((u_int32_t*)(r0.data))[0];
				return( c->result );
			}
				base_stack_pos=pop();
			p= (u_int32_t*) pop();
			r_push(&r0);
			break;
	case oARG:	x=*p++;
			if( x>0 )
				fn_getargsval(&r0,x);
			else	runtime("Invalid argument $%d",x);
			r_push(&r0);
			break;
	case oBIN:	x=*p++;
			fn_getsval_n=*p++;
			push( ((int(*)(struct event_context_s*,int(*)(struct register_s*)))(x))(c,fn_getsval) );
			push(LTI);
			break;
	case oFRC:	x=*p++;
			fn_getsval_n=*p++;
			push( force_do(x,c,fn_getsval) );
			push(LTI);
			break;
	case oPRO:	x=(u_int32_t)p;
			x+= (1+*p++) * sizeof(u_int32_t);
//printf("%p PRO %x [%x %x %x]\n",p-2,x,((u_int32_t*)x)[0],((u_int32_t*)x)[1],((u_int32_t*)x)[2]);
			r_pop(&r0);
			r_imm(&r0);
			{ struct { struct event_handler_s h;
				   u_int32_t d[4];
				 } ph;
#ifdef DEBUG_TRACE
				strcpy(ph.h.op_name+MEDUSA_OPNAME_MAX+1,h->op_name+MEDUSA_OPNAME_MAX);
				ph.h.op_name[MEDUSA_OPNAME_MAX]='#';
#endif
				ph.h.handler=execute_handler;
				ph.h.local_vars=h->local_vars;
				ph.d[0]=oJSR; ph.d[1]=x; ph.d[2]=oRET;
				push( comm_proc(r_int(&r0),&(ph.h)) );
				push(LTI);
			}
			break;
#ifndef DEBUG_TRACE
	default:	runtime("Illegal instruction 0x%08x 0x%08x 0x%08x",cmd,p[0],p[1]);
#else
	default:	runtime("Illegal instruction [%s] 0x%08x 0x%08x 0x%08x",h->op_name+DT_POS_MAX,cmd,p[0],p[1]);
#endif
	}
}
}

int execute_init( void )
{
	stack_size=256;
	stack_pos=0;
	if( (stack=malloc(stack_size*sizeof(u_int32_t)))==NULL )
		return(init_error(Out_of_memory));
	return(0);
}

