
#include "constable.h"
#include "object.h"

static struct class_s *classes=NULL;

struct class_s *get_class_by_name( char *name )
{ struct class_s *c;
	for(c=classes;c!=NULL;c=c->next)
		if( !strncmp(c->m.name,name,MEDUSA_CLASSNAME_MAX) )
			return(c);
	return(NULL);
}

struct class_s *get_class_by_id( u_int16_t id )
{ struct class_s *c;
	for(c=classes;c!=NULL;c=c->next)
		if( c->m.classid==id )
			return(c);
	return(NULL);
}

struct class_s *add_class( struct medusa_class_s *mc, struct medusa_attribute_s *a )
{ struct class_s *c,**p;
  int i,l;
  struct medusa_attribute_s *ci;
	if( a==NULL )	a=(struct medusa_attribute_s *)(mc+1);
	for(i=0;a[i].type!=MED_TYPE_END;i++);
	i++;
	l=sizeof(struct class_s)+i*sizeof(struct medusa_attribute_s);
	if( (c=malloc(l))==NULL )
		return(NULL);
	c->m=*mc;
	memcpy(c->attr,a,i*sizeof(struct medusa_attribute_s));
	if( (ci=get_attribute(c,"cinfo"))!=NULL )
	{	c->cinfo_offset=ci->offset;
		c->cinfo_size=(ci->length)/sizeof(u_int32_t);
	}
	else
		c->cinfo_offset=c->cinfo_size=0;
	c->cinfo_mask=0;
	c->set=NULL;
	if( (ci=get_attribute(c,"vs"))!=NULL )
	{	c->vs_offset=ci->offset;
		c->vs_size=(ci->length)/sizeof(u_int32_t);
	}
	else
		c->vs_offset=c->vs_size=0;
	if( (ci=get_attribute(c,"event"))!=NULL )
	{	c->event_offset=ci->offset;
		c->event_size=ci->length;
	}
	else
		c->event_offset=c->event_size=0;
	for(p=&classes;(*p)!=NULL;p=&((*p)->next))
		if( !strncmp((*p)->m.name,c->m.name,MEDUSA_CLASSNAME_MAX) )
		{	free(c);
			return(NULL);
		}
	c->next=*p;
	*p=c;
	return(c);
}

int class_alloc_cinfo( struct class_s *c )
{ int i;
	for(i=0;i<c->cinfo_size && i<sizeof(c->cinfo_mask)*8;i++)
	{	if( !(c->cinfo_mask & (1<<i)) )
		{	c->cinfo_mask|= 1<<i;
			return(i);
		}
	}
	return(-1);
}

int class_add_sethandler( struct class_s *c, int(*handler)(struct object_s *) )
{ struct class_set_s *s,**p;
	if( (s=malloc(sizeof(struct class_set_s)))==NULL )
		return(-1);
	s->handler=handler;
	for(p=&(c->set);*p!=NULL;p=&((*p)->next));
	s->next=NULL;
	*p=s;
	return(0);
}

int object_do_sethandler( struct object_s *o )
{ struct class_set_s *s;
	for(s=o->class->set;s!=NULL;s=s->next)
		s->handler(o);
	return(0);
}

struct medusa_attribute_s *get_attribute( struct class_s *c, char *name )
{ struct medusa_attribute_s *a;
	if( c==NULL )	return(NULL);
	for(a=c->attr;a->type!=MED_TYPE_END;a++)
		if( !strncmp(a->name,name,MEDUSA_ATTRNAME_MAX) )
			return(a);
	return(NULL);
}

int object_is_same( struct object_s *a, struct object_s *b )
{ struct class_s *c;
  struct medusa_attribute_s *t;
	if( (c=a->class) != b->class )
		return(0);
	for(t=c->attr;t->type!=MED_TYPE_END;t++)
//{if( (t->type & MED_TYPE_PRIMARY_KEY) )
//printf("PK \"%s\" %d ? %d\n",t->name,
//*((int*)(a->data+t->offset)),
//*((int*)(b->data+t->offset)) );
		if( (t->type & MED_TYPE_PRIMARY_KEY)
		     && memcmp(a->data+t->offset,b->data+t->offset,t->length) )
			return(0);
//}
	return(1);
}


int object_clear_event( struct object_s *o )
{
	if( o->class->event_size<=0 )
		return(-1);
	memset(o->data + o->class->event_offset,0,o->class->event_size);
	return(0);
}

int object_add_event( struct object_s *o, event_mask_t *e )
{ int i,n;
  char *f,*t;
	if( o->class->event_size<=0 )
		return(-1);
	n=o->class->event_size;
	if( n>sizeof(event_mask_t) )
		n=sizeof(event_mask_t);
	f=(char *)e;
	t=o->data + o->class->event_offset;
	for(i=0;i<n;i++)
		t[i] |= f[i];
	return(0);
}

int object_set_inherit( struct object_s *o )
{ int n;
  char *t;
	if( o->class->event_size<=0 )
		return(-1);
	n=o->class->event_size;
	t=o->data + o->class->event_offset;
	t[n-1] |= 0x80;
	return(0);
}

int object_clr_inherit( struct object_s *o )
{ int n;
  char *t;
	if( o->class->event_size<=0 )
		return(-1);
	n=o->class->event_size;
	t=o->data + o->class->event_offset;
	t[n-1] &= 0x7f;
	return(0);
}

