
#include "rbac.h"
#include "../constable.h"
#include "../object.h"
#include "../subject.h"
#include "../space.h"
#include "../procmem.h"

#define	SYS_RBACADM	250

/* ---------- rbac tree ---------- */

static int rbac_add_vs( struct tree_s *p, int recursive, int which, const vs_t *vs )
{
	if( which!=AT_MEMBER )
		return(-1);
	if( !recursive )
		vs_add(vs,((struct rbac_role_s*)p)->vs);
	else
		vs_add(vs,((struct rbac_role_s*)p)->recursive_vs);
	return(0);
}

static int rbac_add_vs_user( struct tree_s *p, int recursive, int which, const vs_t *vs )
{	
	if( which!=AT_MEMBER )
		return(-1);
	vs_add(vs,((struct rbac_user_s*)p)->vs);
	return(0);
}

static int rbac_add_vs_ruser( struct tree_s *p, int recursive, int which, const vs_t *vs )
{
	if( which!=AT_MEMBER )
		return(-1);
	if( recursive )
	{	vs_add(vs,((struct rbac_rusers_s*)p)->recursive_vs);
//		rbac_all_users_add_vs(vs);
	}
	vs_add(vs,((struct rbac_rusers_s*)p)->vs);
	return(0);
}

static int rbac_add_vs_perm( struct tree_s *p, int recursive, int which, const vs_t *vs )
{	
	if( which!=AT_MEMBER )
		return(-1);
	vs_add(vs,((struct rbac_perm_s*)p)->vs);
	return(0);
}

static int rbac_add_vs_rrole( struct tree_s *p, int recursive, int which, const vs_t *vs )
{
	if( which!=AT_MEMBER )
		return(-1);
	if( recursive )
	{	vs_add(vs,((struct rbac_rrole_s*)p)->recursive_vs);
//		rbac_all_roles_add_vs(vs);
	}
	vs_add(vs,((struct rbac_rrole_s*)p)->vs);
	return(0);
}

static int rbac_add_vs_rbac( struct tree_s *p, int recursive, int which, const vs_t *vs )
{
//	if( which!=AT_MEMBER )
		return(-1);
}

static int node_init_perm( struct tree_s *t )
{
	vs_clear( ((struct rbac_perm_s*)t)->vs );
	return(0);
}

static tree_type_t tree_type_perm={
			"rbac_permission",
			sizeof(struct rbac_perm_s),
			NULL,
			node_init_perm,
			&tree_type_perm,
			rbac_add_vs_perm,
			};

static int node_init_role( struct tree_s *t )
{ // int i;
//	for(i=0;i<NR_ACCESS_TYPES;i++)
	{	vs_clear(((struct rbac_role_s*)t)->vs);
		vs_clear(((struct rbac_role_s*)t)->recursive_vs);
	}
	return(0);
}

static tree_type_t tree_type_ROLE={
			"rbac_role",
			sizeof(struct rbac_role_s),
			NULL,
			node_init_role,
			&tree_type_ROLE,
			rbac_add_vs,
			};
static tree_type_t tree_type_role={
			"rbac_role",
			sizeof(struct rbac_role_s),
			NULL,
			node_init_role,
			&tree_type_role,
			rbac_add_vs,
			};

static int node_init_rrole( struct tree_s *t )
{
	vs_clear( ((struct rbac_rrole_s*)t)->vs );
	return(0);
}

static tree_type_t tree_type_rROLE={
			"rbac_ROLES",
			sizeof(struct rbac_rrole_s),
			NULL,
			node_init_rrole,
			&tree_type_ROLE,
			rbac_add_vs_rrole,
			};
static tree_type_t tree_type_rrole={
			"rbac_roles",
			sizeof(struct rbac_rrole_s),
			NULL,
			node_init_rrole,
			&tree_type_role,
			rbac_add_vs_rrole,
			};

static int node_init_user( struct tree_s *t )
{
	((struct rbac_user_s*)t)->user=rbac_user_find(t->name);
	return(0);
}

static tree_type_t tree_type_user={
			"rbac_user",
			sizeof(struct rbac_user_s),
			NULL,
			node_init_user,
			&tree_type_user,
			rbac_add_vs_user,
			};

static int node_init_rusers( struct tree_s *t )
{
	vs_clear( ((struct rbac_rusers_s*)t)->vs );
	return(0);
}

static tree_type_t tree_type_rusers={
			"rbac_users",
			sizeof(struct rbac_rusers_s),
			NULL,
			node_init_rusers,
			&tree_type_user,
			rbac_add_vs_ruser,
			};

static int node_init_rbac( struct tree_s *t )
{
	return(0);
}

static tree_type_t tree_type_rbac={
			"rbac",
			sizeof(struct tree_s),
			NULL,
			node_init_rbac,
			NULL,
			rbac_add_vs_rbac,
			};

static int cinfo;
static struct class_s *proc_class;
static int uid_offset;
static int uid_size;
static int vs_offset;
static int vss_offset;
static int vsr_offset;
static int vsw_offset;
static int syscall_offset;
static int sc_syscall;
static int sc_arg0;
static int sc_arg1;
static int sc_arg2;
static int sc_arg3;
static int sc_arg4;

struct rbac_rusers_s *r_user;
struct rbac_rrole_s *r_role,*r_ROLE;
struct rbac_perm_s *r_perm;

static int set_handler_g(struct event_handler_s *h, struct event_context_s *c );

static struct event_handler_s set_handler_d={
			"",
			set_handler_g,
			NULL,
};
static struct event_type_s *set_type=NULL;

static int handler_syscall_f(struct event_handler_s *h, struct event_context_s *c )
{ u_int32_t a;
  char arg0[128];
  char arg1[128];
  char arg2[128];
  char arg3[128];
	if( ((*((u_int16_t*)(c->operation.data + sc_syscall)))&0x0000ffff)
			!=SYS_RBACADM )
	{	c->result=RESULT_UNKNOWN;
		return(0);
	}
	a=*((u_int32_t*)(c->operation.data + sc_arg0));
	procmem_readstr(&(c->subject),a,arg0,128);
	a=*((u_int32_t*)(c->operation.data + sc_arg1));
	procmem_readstr(&(c->subject),a,arg1,128);
	a=*((u_int32_t*)(c->operation.data + sc_arg2));
	procmem_readstr(&(c->subject),a,arg2,128);
	a=*((u_int32_t*)(c->operation.data + sc_arg3));
	procmem_readstr(&(c->subject),a,arg3,128);
	a=*((u_int32_t*)(c->operation.data + sc_arg4));
//printf("ZZZijem: \"%s\" \"%s\" \"%s\" \"%s\" 0x%x\n",arg0,arg1,arg2,arg3,a);

	if( rbac_adm(&(c->subject),arg0,arg1,arg2,arg3)<0 )
		procmem_writestr(&(c->subject),a,errstr);
	else
		procmem_writestr(&(c->subject),a,"OK");
//printf("ZZZijem2: \"%s\"\n",errstr);
	c->result=RESULT_SKIP;
	return(0);
}

static struct event_handler_s handler_syscall={
			"",
			handler_syscall_f,
			NULL,
};
static struct event_type_s *sys_type=NULL;


static int rbac_get_subrole_vs( vs_t *vs, struct role_s *role )
{ struct rbac_role_s *r;
  struct hierarchy_s *h;
	
	for(r=(struct rbac_role_s *)find_one((struct tree_s *)r_ROLE,role->name);r!=NULL;r=(struct rbac_role_s *)(r->tree.alt))
		vs_add(r->recursive_vs,vs);
	for(h=role->sub;h!=NULL;h=h->next_sub)
	{	rbac_get_subrole_vs(vs,h->sub_role);
	}
	return(0);
}

static int rbac_get_suprole_vs( vs_t *vs, struct role_s *role )
{ struct rbac_role_s *r;
  struct hierarchy_s *h;
	
	for(r=(struct rbac_role_s *)find_one((struct tree_s *)r_role,role->name);r!=NULL;r=(struct rbac_role_s *)(r->tree.alt))
		vs_add(r->recursive_vs,vs);
	for(h=role->sup;h!=NULL;h=h->next_sup)
	{	rbac_get_suprole_vs(vs,h->sup_role);
	}
	return(0);
}

int rbac_get_role_vs( vs_t *vs, struct role_s *role )
{ struct rbac_role_s *r;
	
	vs_clear(vs);
	vs_add(r_role->recursive_vs,vs);
	vs_add(r_ROLE->recursive_vs,vs);
	for(r=(struct rbac_role_s *)find_one((struct tree_s *)r_role,role->name);r!=NULL;r=(struct rbac_role_s *)(r->tree.alt))
	{	vs_add(r->vs,vs);
		vs_add(r->recursive_vs,vs);
	}
	for(r=(struct rbac_role_s *)find_one((struct tree_s *)r_ROLE,role->name);r!=NULL;r=(struct rbac_role_s *)(r->tree.alt))
	{	vs_add(r->vs,vs);
		vs_add(r->recursive_vs,vs);
	}
	rbac_get_subrole_vs(vs,role);
	rbac_get_suprole_vs(vs,role);
	return(0);
}

int rbac_get_user_vs( vs_t *vs, struct user_s *user )
{ struct rbac_user_s *u;
	
	vs_clear(vs);
	vs_add(r_user->recursive_vs,vs);
	for(u=(struct rbac_user_s *)find_one((struct tree_s *)r_user,user->name);u!=NULL;u=(struct rbac_user_s *)(u->tree.alt))
	{	vs_add(u->vs,vs);
	}
	return(0);
}


static int set_handler( struct object_s *o )
{ u_int32_t *s_ci;
  vs_t *vs,*vss,*vsr,*vsw;
  struct user_assignment_s *u;
	s_ci= (u_int32_t *)(o->data + o->class->cinfo_offset);
	vs=(vs_t*)(o->data + vs_offset);
	vss=(vs_t*)(o->data + vss_offset);
	vsr=(vs_t*)(o->data + vsr_offset);
	vsw=(vs_t*)(o->data + vsw_offset);
//printf("ZZZrbac set_handler pred vs=%x vss=%x, vsr=%x, vsw=%x\n",*vs,*vss,*vsr,*vsw);

	object_add_event(o,&(set_type->mask));

	((u_int32_t *)(o->data + syscall_offset))[SYS_RBACADM>>5]|=
			1<<(SYS_RBACADM&0x1f);

	if( s_ci[cinfo]==0 )
		return(0);
	
	u=((struct user_s *)(s_ci[cinfo]))->ua;	/* ma byt session */

	for(;u!=NULL;u=u->next_role)
	{ vs_t pvs[MAX_VS_BITS/32];
		vs_add(u->role->vs[AT_MEMBER],vs);
		rbac_get_role_vs(pvs,u->role);
		vs_add(pvs,vs);
//		vs_add(u->role->recursive_vs_sup[AT_MEMBER],vs);
//		vs_add(u->role->recursive_vs_sub[AT_MEMBER],vs);
		vs_add(u->role->vs[AT_SEE],vss);
//		vs_add(u->role->recursive_vs_sup[AT_SEE],vss);
//		vs_add(u->role->recursive_vs_sub[AT_SEE],vss);
		vs_add(u->role->vs[AT_RECEIVE],vsr);
//		vs_add(u->role->recursive_vs_sup[AT_RECEIVE],vsr);
//		vs_add(u->role->recursive_vs_sub[AT_RECEIVE],vsr);
		vs_add(u->role->vs[AT_SEND],vsw);
//		vs_add(u->role->recursive_vs_sup[AT_SEND],vsw);
//		vs_add(u->role->recursive_vs_sub[AT_SEND],vsw);
	}
//printf("ZZZrbac set_handler vs=%x vss=%x, vsr=%x, vsw=%x\n",*vs,*vss,*vsr,*vsw);
	return(0);
}

static int set_handler_g(struct event_handler_s *h, struct event_context_s *c )
{ u_int32_t *s_ci;
  uid_t uid=(uid_t)-1;
  vs_t *vs,*vss,*vsr,*vsw;
	s_ci= (u_int32_t *)(c->subject.data + c->subject.class->cinfo_offset);

	switch( uid_size )
	{	case 1:	uid=(uid_t)(*((char*)(c->subject.data + uid_offset)));
			break;
		case 2:	uid=(uid_t)(*((u_int16_t*)(c->subject.data + uid_offset)));
			break;
		case 4:	uid=(uid_t)(*((u_int32_t*)(c->subject.data + uid_offset)));
			break;
		case 8:	uid=(uid_t)(*((u_int64_t*)(c->subject.data + uid_offset)));
			break;
	}
	s_ci[cinfo]=(u_int32_t)rbac_user_find_by_uid(uid);
//if( s_ci[cinfo]!=0 )
//printf("ZZZ RBAC user=%s\n",((struct user_s *)(s_ci[cinfo]))->name);
//else
//printf("ZZZ RBAC user none (uid=%d)\n",uid);

	vs=(vs_t*)(c->subject.data + vs_offset);
	vss=(vs_t*)(c->subject.data + vss_offset);
	vsr=(vs_t*)(c->subject.data + vsr_offset);
	vsw=(vs_t*)(c->subject.data + vsw_offset);
	vs_clear(vs);
	vs_clear(vss);
	vs_clear(vsr);
	vs_clear(vsw);
	object_clear_event(&(c->subject));
	object_do_sethandler(&(c->subject));
	c->result=RESULT_OK;
	return(0);
}

static int rbac_get_vs( vs_t *vs, int n, struct object_s *o )
{ u_int32_t *s_ci;
  struct user_assignment_s *u;
	if( o->class!=proc_class )
		return(-1);
	s_ci= (u_int32_t *)(o->data + o->class->cinfo_offset);
	if( s_ci[cinfo]==0 )
		return(0);
	for(u=((struct user_s *)(s_ci[cinfo]))->ua;u!=NULL;u=u->next_role)
	{	vs_add( u->role->vs[n] , vs );
		if( n==AT_MEMBER )
		{ vs_t pvs[MAX_VS_BITS/32];
			rbac_get_role_vs(pvs,u->role);
			vs_add(pvs,vs);
		}
//		vs_add( u->role->recursive_vs_sup[n] , vs );
//		vs_add( u->role->recursive_vs_sub[n] , vs );
	}
	return(0);
}




int rbac_init( void )
{ struct medusa_attribute_s *a;
	if( (proc_class=get_class_by_name("process"))==NULL )
		if( (proc_class=get_class_by_name("proc"))==NULL )
			return(init_error("RBAC: Class 'proc' don't exist"));
	if( (cinfo=class_alloc_cinfo(proc_class))<0 )
		return(init_error("RBAC: No free cinfo"));

	if( (a=get_attribute(proc_class,"uid"))==NULL )
		return(init_error("RBAC: Class 'proc' don't have attribute 'vs'"));
	uid_offset=a->offset;
	uid_size=a->length;
	if( (a=get_attribute(proc_class,"vs"))==NULL )
		return(init_error("RBAC: Class 'proc' don't have attribute 'vs'"));
	vs_offset=a->offset;
	if( (a=get_attribute(proc_class,"vss"))==NULL )
		return(init_error("RBAC: Class 'proc' don't have attribute 'vss'"));
	vss_offset=a->offset;
	if( (a=get_attribute(proc_class,"vsr"))==NULL )
		return(init_error("RBAC: Class 'proc' don't have attribute 'vsr'"));
	vsr_offset=a->offset;
	if( (a=get_attribute(proc_class,"vsw"))==NULL )
		return(init_error("RBAC: Class 'proc' don't have attribute 'vsw'"));
	vsw_offset=a->offset;
	if( (a=get_attribute(proc_class,"syscall"))==NULL || a->length < SYS_RBACADM/8 )
		return(init_error("RBAC: Class 'proc' don't have attribute 'syscall'"));
	syscall_offset=a->offset;

	if( (set_type=event_type_find_name("setuid",proc_class,NULL,proc_class))==NULL )
		return(init_error("RBAC: No 'setuid' operation!"));

	if( (sys_type=event_type_find_name("syscall",proc_class,NULL,proc_class))==NULL )
		return(init_error("RBAC: No 'syscall' operation!"));
	if( (a=get_attribute(&(sys_type->operation_class),"syscall"))==NULL || a->length!=sizeof(u_int16_t) )
		return(init_error("RBAC: Class 'syscall' don't have attribute 'syscall'"));
	sc_syscall=a->offset;
	if( (a=get_attribute(&(sys_type->operation_class),"arg0"))==NULL || a->length!=sizeof(u_int32_t) )
		return(init_error("RBAC: Class 'syscall' don't have attribute 'arg0'"));
	sc_arg0=a->offset;
	if( (a=get_attribute(&(sys_type->operation_class),"arg1"))==NULL || a->length!=sizeof(u_int32_t) )
		return(init_error("RBAC: Class 'syscall' don't have attribute 'arg1'"));
	sc_arg1=a->offset;
	if( (a=get_attribute(&(sys_type->operation_class),"arg2"))==NULL || a->length!=sizeof(u_int32_t) )
		return(init_error("RBAC: Class 'syscall' don't have attribute 'arg2'"));
	sc_arg2=a->offset;
	if( (a=get_attribute(&(sys_type->operation_class),"arg3"))==NULL || a->length!=sizeof(u_int32_t) )
		return(init_error("RBAC: Class 'syscall' don't have attribute 'arg3'"));
	sc_arg3=a->offset;
	if( (a=get_attribute(&(sys_type->operation_class),"arg4"))==NULL || a->length!=sizeof(u_int32_t) )
		return(init_error("RBAC: Class 'syscall' don't have attribute 'arg4'"));
	sc_arg4=a->offset;

	if( class_add_sethandler(proc_class,set_handler)<0 )
		return(init_error("RBAC: Can't add sethandler"));
	if( register_event_handler(&set_handler_d,set_type,ALL_OBJ,NULL)<0 )
		return(init_error("RBAC: Can't register handler"));
	if( register_event_handler(&handler_syscall,sys_type,ALL_OBJ,NULL)<0 )
		return(init_error("RBAC: Can't register syscall handler"));
	if( subject_register(rbac_get_vs)<0 )
		return(init_error("RBAC: Can't register subject"));

	if( register_tree_type(&tree_type_rbac)<0 )
		return(init_error("RBAC: Can't register tree type"));
	tree_type_rbac.child_type=&tree_type_rusers;
	r_user=(struct rbac_rusers_s*)create_path("rbac/user");

	tree_type_rbac.child_type=&tree_type_rrole;
	r_role=(struct rbac_rrole_s*)create_path("rbac/role");

	tree_type_rbac.child_type=&tree_type_rROLE;
	r_ROLE=(struct rbac_rrole_s*)create_path("rbac/ROLE");

	tree_type_rbac.child_type=&tree_type_perm;
	r_perm=(struct rbac_perm_s*)create_path("rbac/permission");

	return(0);
}

