
#include "constable.h"
#include "tree.h"
#include <regex.h>

#define	DEFAULT_PATH	"fs"

static struct tree_type_s global_root_type={
		"/",
		sizeof(struct tree_s),
		NULL,
		NULL,
		NULL,
		NULL
};

struct tree_s global_root={NULL,NULL,NULL,NULL,NULL,&global_root_type,
				NULL,NULL,NULL,NULL,0,0};

/* ---------- regexps ---------- */

static int isreg( char *name )
{ int unbs=0;
  char *n=name;
	while( *name )
	{	switch (*name)
		{	case '(':
			case ')':
			case '[':
			case ']':
			case '^':
			case '$':
			case '.':
			case '*':
			case '+':
			case '?':
			case '{':
			case '}':
			case '|':
				return 1;
			case '\\':
				if (
				    (name[1] >= 'A' && name[1] <= 'Z') ||
				    (name[1] >= 'a' && name[1] <= 'z') ||
				    name[1]==0
				    )
					return 1;
				unbs=1;
				name++;
		}
		name++;
	}
	if( unbs )
	{ int i=0;
		while( n[i] )
		{	if( n[i]=='\\' )
				i++;
			n[0]=n[i];
			n++;
		}
		n[0]=0;
	}
	return 0;
}

static void *regcompile( char *reg )
{ int l=strlen(reg);
  char tmp[l+3];
  regex_t *r;
	if( !strcmp(reg,".*") )
		return((void*)1);	/* speed up .* */
  	if( (r=malloc(sizeof(regex_t)))==NULL )
		return(NULL);
	tmp[0]='^';
	strcpy(tmp+1,reg);
	tmp[l+1]='$';
	tmp[l+2]=0;
	if( regcomp(r, tmp, REG_EXTENDED|REG_NOSUB )!=0 )
	{	error("Error in regexp '%s'", reg);
		free(r);
		return(NULL);
	}
	return(r);
}

static int regcmp( regex_t *r, char *name )
{
	if( r==(void*)1 )
		return(0);	/* speed up .* */
	if( regexec(r,name,0,NULL,0)==0 )
		return(0);
	return(-1);
}

/* ---------- tree ---------- */

static int pathcmp( char *name, char **path )
{ char *b;
	b=*path;
	while( *name!=0 && *name==*b )
	{	name++; b++;	}
	if( *name==0 && (*b==0 || *b=='/') )
	{	*path=b;
		return(0);
	}
	if( *b>*name )
		return(1);
	return(-1);
}

static struct tree_s *create_one( struct tree_s *base, char **name )
{ struct tree_s *p;
  struct tree_type_s *type,*ctype;
  char *n;
  int l=0;
  
	while( **name=='/' )	(*name)++;
	if( **name==0 )		return(base);
	for(p=base->child;p!=NULL;p=p->next)
		if( !pathcmp(p->name,name) )
			return(p);
	for(p=base->reg;p!=NULL;p=p->next)
		if( !pathcmp(p->name,name) )
			return(p);
	n=*name;
	while( *n!=0 && *n!='/' )
	{	l++;
		n++;
	}
	if( (type=base->type->child_type)==NULL )
		type=base->type;
	if( (ctype=type->child_type)==NULL )
		ctype=type;
	if( (p=malloc(type->size+ctype->size+l+1+4))==NULL )
		return(NULL);
	p->type=type;
	p->name=((char*)p)+(type->size+ctype->size)+4;
	strncpy(p->name,*name,l);
	p->name[l]=0;
	*name=n;
	p->parent=base;
	p->child=NULL;
	p->reg=NULL;
	p->primary_space=NULL;
	p->recursive_primary_space=base->recursive_primary_space;
	p->event=0;
	p->recursive_event=base->recursive_event;
	if( !isreg(p->name) )
	{ struct tree_s *a;
		p->next=base->child;
		base->child=p;
		p->extra=NULL;
		for(a=base->reg;a!=NULL && regcmp(a->extra,p->name);a=a->next);
		p->alt=a;
	}
	else
	{ struct tree_s *a;
		p->next=base->reg;
		base->reg=p;
		p->extra=regcompile(p->name);
		for(p->alt=p->next;p->alt!=NULL && strcmp(p->alt->name,".*");
				p->alt=p->alt->next);
		for(a=base->child;a!=NULL;a=a->next)
			if( a->alt==p->alt && !regcmp(p->extra,a->name) )
				a->alt=p;
	}
	p->type->init(p);
	p->reg=(struct tree_s*)(((char*)p)+(p->type->size));
	p->reg->type=ctype;
	p->reg->name=((char*)p)+(type->size+ctype->size);
	strcpy(p->reg->name,".*");
	p->reg->extra=regcompile(p->reg->name);
	p->reg->parent=p;
	p->reg->child=NULL;
	p->reg->reg=NULL;
	p->reg->next=NULL;
	p->reg->alt=NULL;
	p->reg->primary_space=NULL;
	p->reg->recursive_primary_space=p->recursive_primary_space;
	p->reg->event=0;
	p->reg->recursive_event=p->recursive_event;
	p->reg->type->init(p->reg);
	return(p);
}

int register_tree_type( tree_type_t *type )
{ char *path;
	global_root.type->child_type=type;	/* !!! POZOR !!! */
	path=type->name;
	if( create_one(&global_root,&path)==NULL )
		return(-1);
	global_root.type->child_type=NULL;
	return(0);
}

struct tree_s *create_path( char *path )
{ struct tree_s *p;
	p=&global_root;
#ifdef DEFAULT_PATH
	if( *path=='/' )
	{ char *d=DEFAULT_PATH;
		while( *d!=0 && (p=create_one(p,&d))!=NULL );
		if( p==NULL )
			return(p);
	}
#endif
	while( *path!=0 && (p=create_one(p,&path))!=NULL );
	return(p);
}

struct tree_s *find_one( struct tree_s *base, char *name )
{ struct tree_s *p;
	while( *name=='/' )	(name)++;
	if( *name==0 )		return(base);
	for(p=base->child;p!=NULL;p=p->next)
		if( !strcmp(p->name,name) )
			return(p);
	for(p=base->reg;p!=NULL;p=p->next)
		if( !regcmp(p->extra,name) )
			return(p);
/* ... zeby este cyklus base=base->alt ??? a lepsie to retazit cez alt */
	if( base->child==NULL && base->reg==NULL )
		return(base);
	return(NULL);
}

int tree_is_offspring( struct tree_s *offspring, struct tree_s *ancestor )
{
	while( offspring!=NULL )
	{	if( offspring==ancestor )
			return(1);
		offspring=offspring->parent;
	}
	return(0);
}

/* ----------------------------------- */

#if 0
int path_add_vs( char *path, int recursive, int which, const vs_t *vs )
{ struct tree_s *p;
	if( (p=create_path(path))==NULL )
		return(-1);
	return(p->type->add_vs(p,recursive,which,vs));
}
#endif

/* ----------------------------------- */

char *tree_get_path( struct tree_s *t )
{ static char buf[4096];
  static int pos=0;
	if( t->parent==NULL )
	{	buf[(pos=0)]=0;
		return(buf);
	}
	tree_get_path(t->parent);
	buf[pos++]='/';
	strncpy(buf+pos,t->name,4095-pos);
	pos+=strlen(t->name);
	if( pos>4095 )	pos=4095;
	return(buf);
}

